from networkx.algorithms import community
from community import community_louvain
import pandas as pd
import networkx as nx
import matplotlib.pyplot as plt
import matplotlib.colors as mcolors
import os
from . import network_analysis
from . import simple_network

def open_network(excel_file_path):

    df = pd.read_excel(excel_file_path)

    # Create a graph
    G = nx.Graph()

    # Add edges from the DataFrame
    edges = df.values  # Assuming the columns are named "Node1" and "Node2"
    G.add_edges_from(edges)

    return G

def save_figure(directory_path, identifier):
    file_name = f"Network {identifier}"
    full_path = os.path.join(directory_path, file_name)
    plt.savefig(full_path)

def weighted_network(excel_file_path):
    """creates a network where the edges have weights proportional to the number of connections they make between the same structure"""
    # Read the Excel file into a pandas DataFrame
    master_list = read_excel_to_lists(excel_file_path)

    # Create a graph
    G = nx.Graph()

    # Create a dictionary to store edge weights based on node pairs
    edge_weights = {}

    nodes_a = master_list[0]
    nodes_b = master_list[1]

    # Iterate over the DataFrame rows and update edge weights
    for i in range(len(nodes_a)):
        node1, node2 = nodes_a[i], nodes_b[i]
        edge = (node1, node2) if node1 < node2 else (node2, node1)  # Ensure consistent order
        edge_weights[edge] = edge_weights.get(edge, 0) + 1

    # Add edges to the graph with weights
    for edge, weight in edge_weights.items():
        G.add_edge(edge[0], edge[1], weight=weight)

    return G, edge_weights

def get_edge_weights(G):
    edge_weights = {}
    weight_G = G.edges(data = True)

    for item in weight_G:
        nodea = item[0]
        nodeb = item[1]
        weight_dict = item[2]
        try:
            weight = weight_dict['weight']
        except KeyError:
            weight = 1

        key = (nodea, nodeb)
        edge_weights[key] = weight

    return edge_weights

def read_excel_to_lists(file_path, sheet_name=0):
    """Convert a pd dataframe to lists"""
    # Read the Excel file into a DataFrame without headers
    df = pd.read_excel(file_path, header=None, sheet_name=sheet_name)

    df = df.drop(0)

    # Initialize an empty list to store the lists of values
    data_lists = []

    # Iterate over each column in the DataFrame
    for column_name, column_data in df.items():
        # Convert the column values to a list and append to the data_lists
        data_lists.append(column_data.tolist())

    master_list = [[], [], []]


    for i in range(0, len(data_lists), 3):

        master_list[0].extend(data_lists[i])
        master_list[1].extend(data_lists[i+1])

        try:
            master_list[2].extend(data_lists[i+2])
        except IndexError:
            pass

    return master_list




def louvain_mod_solo(G, edge_weights = None, identifier = None, directory_path = None):

    if type(G) == str:
        G, edge_weights = weighted_network(G)

    if edge_weights is None:
        edge_weights = get_edge_weights(G)

    # Assuming G is your NetworkX graph
    # Louvain community detection
    partition = community_louvain.best_partition(G)

    # Print modularity
    modularity = community_louvain.modularity(partition, G)
    number_of_nodes = G.number_of_nodes()
    print("Modularity:", modularity)
    num_components = nx.number_connected_components(G)

    if directory_path:
        with open(f'{directory_path}/{identifier} network stats.txt', 'a') as f:
            f.write(f'\nlouvain full graph modularity: {modularity} constructed from {num_components} components containing {number_of_nodes} nodes')
        f.close()

    louvain_graph(edge_weights, identifier, directory_path)

def louvain_mod(G, edge_weights=None, identifier=None, geometric = False, geo_info = None, directory = None):

    if type(G) == list:
        num_edge = len(G[0])
    else:
        num_edge = None


    if type(G) == str or type(G) == list:
        G, edge_weights = network_analysis.weighted_network(G)

    if edge_weights is None:
        edge_weights = get_edge_weights(G)

    connected_components = list(nx.connected_components(G))
    num_nodes = float(G.number_of_nodes())

    print("Number of nodes:", num_nodes)
    print("Number of edges:", num_edge)

    # Calculate the average degree connectivity
    average_degree_connectivity = nx.average_degree_connectivity(G)
    print("Average degree connectivity:", average_degree_connectivity)

    # Calculate the average number of edges attached to a node
    average_edges_per_node = num_nodes/num_edge
    print("Average edges per node:", average_edges_per_node)

    for i, component in enumerate(connected_components):
        # Apply the Louvain community detection on the subgraph
        partition = community_louvain.best_partition(G.subgraph(component))

        # Invert the partition dictionary to get communities
        #communities = {}
        #for node, comm_id in partition.items():
            #communities.setdefault(comm_id, []).append(node)
        #communities = list(communities.values())
        
        # Assign a different color to each community within the component for visualization
        #colors = [mcolors.to_hex(plt.cm.tab10(i / len(connected_components))[:3]) for _ in range(len(communities))]
        
        # Calculate modularity
        modularity = community_louvain.modularity(partition, G.subgraph(component))
        num_nodes = len(component)

        print(f"Louvain modularity for component with {num_nodes} nodes: {modularity}")


    louvain_graph(edge_weights, identifier, directory, geometric, geo_info)

def show_communities_flex(G, master_list, normalized_weights, geo_info = None, geometric=False, directory=None, weighted=True, partition=None, style=0):
    if partition is None:
        partition, normalized_weights = community_partition(master_list, weighted=weighted, style=style)
        print(partition)
    if normalized_weights is None:
        G, edge_weights = network_analysis.weighted_network(master_list)

        if edge_weights is None:
            edge_weights = get_edge_weights(G)
        # Find the maximum and minimum edge weights
        max_weight = max(weight for edge, weight in edge_weights.items())
        min_weight = min(weight for edge, weight in edge_weights.items())

        if max_weight > 1:
            # Normalize edge weights to the range [0.1, 1.0]
            normalized_weights = {edge: 0.1 + 0.9 * ((weight - min_weight) / (max_weight - min_weight)) for edge, weight in edge_weights.items()}
        else:
            normalized_weights = {edge: 0.1 for edge, weight in edge_weights.items()}

        # Add edges to the graph with normalized weights
        for edge, normalized_weight in normalized_weights.items():
            G.add_edge(edge[0], edge[1], weight=normalized_weight)

    if geometric:
        for node in list(G.nodes()):
            if node not in geo_info[0]:
                G.remove_node(node)
                print(f"Removing node {node} from network visualization (no centroid - likely due to downsampling when finding centroids)")

    # Group nodes by their community
    communities = {}
    for node, community_id in partition.items():
        communities.setdefault(community_id, []).append(node)

    # Create a mapping of community IDs to sequential indices
    unique_communities = sorted(set(partition.values()))
    community_to_index = {comm: idx for idx, comm in enumerate(unique_communities)}

    # Prepare colors using the number of unique communities
    colors = [plt.cm.jet(i / len(unique_communities)) for i in range(len(unique_communities))]

    if weighted:
        G = nx.Graph()

        # Add edges to the graph with normalized weights
        for edge, normalized_weight in normalized_weights.items():
            G.add_edge(edge[0], edge[1], weight=normalized_weight)

        if geometric:
            pos, z_pos = simple_network.geometric_positions(geo_info[0], geo_info[1])
            
            # Draw the nodes, coloring them according to their community
            for community_id, nodes in communities.items():
                node_sizes_list = [z_pos[node] for node in nodes]
                nx.draw_networkx_nodes(G, pos, nodelist=nodes, 
                                     node_color=[colors[community_to_index[community_id]]], 
                                     node_size=node_sizes_list, alpha=0.8)

            # Draw edges with normalized weights
            for edge in G.edges():
                normalized_weight = G[edge[0]][edge[1]]['weight']
                nx.draw_networkx_edges(G, pos, edgelist=[edge], width=5 * normalized_weight, edge_color='black')

            nx.draw_networkx_labels(G, pos)

        else:
            pos = nx.spring_layout(G)
            
            # Draw the nodes, coloring them according to their community
            for community_id, nodes in communities.items():
                nx.draw_networkx_nodes(G, pos, nodelist=nodes, 
                                     node_color=[colors[community_to_index[community_id]]], 
                                     node_size=100, alpha=0.8)

            # Draw edges with normalized weights
            for edge in G.edges():
                normalized_weight = G[edge[0]][edge[1]]['weight']
                nx.draw_networkx_edges(G, pos, edgelist=[edge], width=5 * normalized_weight, edge_color='black')

            nx.draw_networkx_labels(G, pos)

    else:
        # Create node color list based on partition and mapping
        node_colors = [colors[community_to_index[partition[node]]] for node in G.nodes()]

        if geometric:
            pos, z_pos = geometric_positions(geo_info[0], geo_info[1])
            node_sizes_list = [z_pos[node] for node in G.nodes()]
            nx.draw(G, pos, with_labels=True, font_color='black', font_weight='bold', 
                   node_size=node_sizes_list, node_color=node_colors, alpha=0.8, font_size=12)
        else:
            pos = nx.spring_layout(G)
            nx.draw(G, pos, with_labels=True, font_color='black', font_weight='bold', 
                   node_size=100, node_color=node_colors, alpha=0.8)

    plt.axis('off')
    plt.show()

    if directory is not None:
        plt.savefig(f'{directory}/network_plot.png')

    return partition, normalized_weights









def community_partition(master_list, weighted = False, style = 0):

    if weighted:
        G, edge_weights = network_analysis.weighted_network(master_list)
        edge_weights = get_edge_weights(G)

    if style == 1 and weighted:

        G = nx.Graph()

        # Find the maximum and minimum edge weights
        max_weight = max(weight for edge, weight in edge_weights.items())
        min_weight = min(weight for edge, weight in edge_weights.items())

        if max_weight > 1:
            # Normalize edge weights to the range [0.1, 1.0]
            normalized_weights = {edge: 0.1 + 0.9 * ((weight - min_weight) / (max_weight - min_weight)) for edge, weight in edge_weights.items()}
        else:
            normalized_weights = {edge: 0.1 for edge, weight in edge_weights.items()}

        # Add edges to the graph with normalized weights
        for edge, normalized_weight in normalized_weights.items():
            G.add_edge(edge[0], edge[1], weight=normalized_weight)

        # Perform Louvain community detection
        partition = community_louvain.best_partition(G)

        return partition, normalized_weights

    elif style == 1:

        edges = list(zip(master_list[0], master_list[1]))

        G = nx.Graph()

        G.add_edges_from(edges)


        # Perform Louvain community detection
        partition = community_louvain.best_partition(G)

        return partition, None

    elif style == 0 and weighted:

        G = nx.Graph()

        # Find the maximum and minimum edge weights
        max_weight = max(weight for edge, weight in edge_weights.items())
        min_weight = min(weight for edge, weight in edge_weights.items())

        if max_weight > 1:
            # Normalize edge weights to the range [0.1, 1.0]
            normalized_weights = {edge: 0.1 + 0.9 * ((weight - min_weight) / (max_weight - min_weight)) for edge, weight in edge_weights.items()}
        else:
            normalized_weights = {edge: 0.1 for edge, weight in edge_weights.items()}

        # Add edges to the graph with normalized weights
        for edge, normalized_weight in normalized_weights.items():
            G.add_edge(edge[0], edge[1], weight=normalized_weight)

        # Detect communities using label propagation
        communities = list(community.label_propagation_communities(G))
        output = {}
        for i, com in enumerate(communities):
            for node in com:
                output[node] = i + 1

        return output, normalized_weights

    elif style == 0:


        edges = list(zip(master_list[0], master_list[1]))

        # Create a graph
        G = nx.Graph()

        # Add edges from the DataFrame
        G.add_edges_from(edges)

        # Detect communities using label propagation
        communities = list(community.label_propagation_communities(G))
        output = {}

        # Detect communities using label propagation
        communities = list(community.label_propagation_communities(G))
        output = {}
        for i, com in enumerate(communities):
            for node in com:
                output[node] = i + 1

        return output, None



def _louvain_mod_solo(G, edge_weights = None):


    if type(G) == str:
        G, edge_weights = weighted_network(G)

    # Louvain community detection
    partition = community_louvain.best_partition(G)

    # modularity
    modularity = community_louvain.modularity(partition, G)

    print(f"Modularity is {modularity}")

    return modularity

def _louvain_mod(G, edge_weights = None):

    if type(G) == str:
        G, edge_weights = weighted_network(G)

    connected_components = list(nx.connected_components(G))
    return_dict = {}

    for i, component in enumerate(connected_components):
        # Apply the Louvain community detection on the subgraph
        partition = community_louvain.best_partition(G.subgraph(component))
        
        # Calculate modularity
        modularity = community_louvain.modularity(partition, G.subgraph(component))
        num_nodes = len(component)
        return_dict[num_nodes] = modularity

        print(f"Louvain modularity for component with {num_nodes} nodes: {modularity}")

    return return_dict

def louvain_graph(edge_weights, identifier, directory_path, geometric, geo_info):
    G = nx.Graph()

    # Find the maximum and minimum edge weights
    max_weight = max(weight for edge, weight in edge_weights.items())
    min_weight = min(weight for edge, weight in edge_weights.items())

    if max_weight > 1:
        # Normalize edge weights to the range [0.1, 1.0]
        normalized_weights = {edge: 0.1 + 0.9 * ((weight - min_weight) / (max_weight - min_weight)) for edge, weight in edge_weights.items()}
    else:
        normalized_weights = {edge: 0.1 for edge, weight in edge_weights.items()}

    # Add edges to the graph with normalized weights
    for edge, normalized_weight in normalized_weights.items():
        G.add_edge(edge[0], edge[1], weight=normalized_weight)

    if geometric:
        for node in list(G.nodes()):
            if node not in geo_info[0]:
                G.remove_node(node)
                print(f"Removing node {node} from network visualization (no centroid - likely due to downsampling when finding centroids)")

    # Perform Louvain community detection
    partition = community_louvain.best_partition(G)

    # Invert the partition dictionary to group nodes by their community
    print(f"louvain_communities_initial: {partition}")
    communities = {}
    for node, community_id in partition.items():
        communities.setdefault(community_id, []).append(node)

    print(f"louvain_communities_after: {partition}")

    # Prepare colors for each community
    unique_communities = set(partition.values())
    colors = [plt.cm.jet(i / len(unique_communities)) for i in range(len(unique_communities))]

    if geometric:

        pos, z_pos  = simple_network.geometric_positions(geo_info[0], geo_info[1])
        #nx.draw(G, pos, with_labels=True, font_color='black', font_weight='bold', node_size = node_sizes_list, node_color = node_color_list, alpha=0.8, font_size = 12)

        # Draw the nodes, coloring them according to their community
        for community_id, nodes in communities.items():
            node_sizes_list = [z_pos[node] for node in nodes]
            nx.draw_networkx_nodes(G, pos, nodelist=nodes, node_color=[colors[community_id]], node_size=node_sizes_list, alpha=0.8)

        # Modify edge drawing to consider normalized weights
        for edge in G.edges():
            normalized_weight = G[edge[0]][edge[1]]['weight']
            
            # Scale the width based on a constant factor (e.g., 5)
            nx.draw_networkx_edges(G, pos, edgelist=[edge], width=5 * normalized_weight, edge_color='black')

        # Optionally, draw node labels
        nx.draw_networkx_labels(G, pos)


    else:
        # Position nodes using the spring layout
        pos = nx.spring_layout(G)

        # Draw the nodes, coloring them according to their community
        for community_id, nodes in communities.items():
            nx.draw_networkx_nodes(G, pos, nodelist=nodes, node_color=[colors[community_id]], node_size=100, alpha=0.8)

        # Modify edge drawing to consider normalized weights
        for edge in G.edges():
            normalized_weight = G[edge[0]][edge[1]]['weight']
            
            # Scale the width based on a constant factor (e.g., 5)
            nx.draw_networkx_edges(G, pos, edgelist=[edge], width=5 * normalized_weight, edge_color='black')

        # Optionally, draw node labels
        nx.draw_networkx_labels(G, pos)

    plt.axis('off')
    if directory_path is not None:
        plt.savefig(f'{directory_path}/community_louvain_network_plot.png')

    plt.show()

def create_directory(directory_name):
    try:
        os.mkdir(directory_name)
        print(f"Directory '{directory_name}' created successfully.")
    except Exception as e:
        pass

def find_threshold(G):
    largest_component = max(nx.connected_components(G), key=len)
    largest_component_subgraph = G.subgraph(largest_component)
    largest_component = float(largest_component_subgraph.number_of_nodes())
    threshold = int(round(0.1 * largest_component))

    return threshold


def remove_small_components(G, threshold):
    """
    Remove connected components with fewer nodes than the threshold from the graph G.

    Parameters:
    - G: A NetworkX graph
    - threshold: An integer, the minimum size required for a component to be retained

    Returns:
    - A modified graph G with small components removed.
    """
    for component in list(nx.connected_components(G)):
        if len(component) < threshold:
            # Remove the nodes of the component from the graph
            G.remove_nodes_from(component)
    return G


def louvain_analysis_5x(G, excel_file, identifier, directory_path):
    print("Performing Louvain Community Detection...")
    df = pd.read_excel(excel_file)
    G, edge_weights = weighted_network(excel_file)
    threshold = find_threshold(G)
    G = remove_small_components(G, threshold)
    louvain_mod(G, identifier, directory_path)
    louvain_graph(edge_weights, identifier, directory_path)
    print("Done")

def louvain_analysis_20x(G, excel_file, identifier, directory_path):
    print("Performing Louvain Community Detection...")
    df = pd.read_excel(excel_file)
    G, edge_weights = weighted_network(excel_file)
    louvain_mod_20x(G, identifier, directory_path)
    louvain_graph(edge_weights, identifier, directory_path)
    print("Done")


if __name__ == "__main__":
    excel_name = input("excel file?: ")
    is_5x = input("Is 5x? (Type Y for 5x): ")
    if is_5x == 'Y':

        df = pd.read_excel(excel_name)
        identifier = "output"
        directory_path = 'output'
        create_directory(directory_path)
        G, edge_weights = weighted_network(excel_name)
        #threshold = find_threshold(G)
        #G = remove_small_components(G, threshold)
        louvain_mod_solo(G, edge_weights, identifier, directory_path)

    else:
        df = pd.read_excel(excel_name)
        identifier = "output"
        directory_path = 'output'
        create_directory(directory_path)
        G, edge_weights = weighted_network(excel_name)
        louvain_mod(G, edge_weights, identifier, directory_path)


import tifffile
import numpy as np
from scipy.ndimage import binary_dilation, distance_transform_edt
from scipy.ndimage import gaussian_filter
from concurrent.futures import ThreadPoolExecutor, as_completed
import cv2
import os
import math
import re
from . import nettracer
import multiprocessing as mp
from skimage.feature import peak_local_max
try:
    import cupy as cp
    import cupyx.scipy.ndimage as cpx
    from cupyx.scipy.ndimage import maximum_filter
except:
    pass


def dilate_3D(tiff_array, dilated_x, dilated_y, dilated_z):
    """Internal method to dilate an array in 3D.
    Arguments are an array,  and the desired pixel dilation amounts in X, Y, Z."""

    def create_circular_kernel(diameter):
        """Create a 2D circular kernel with a given radius.

        Parameters:
        radius (int or float): The radius of the circle.

        Returns:
        numpy.ndarray: A 2D numpy array representing the circular kernel.
        """
        # Determine the size of the kernel
        radius = diameter/2
        size = radius  # Diameter of the circle
        size = int(np.ceil(size))  # Ensure size is an integer
        
        # Create a grid of (x, y) coordinates
        y, x = np.ogrid[-radius:radius+1, -radius:radius+1]
        
        # Calculate the distance from the center (0,0)
        distance = np.sqrt(x**2 + y**2)
        
        # Create the circular kernel: points within the radius are 1, others are 0
        kernel = distance <= radius
        
        # Convert the boolean array to integer (0 and 1)
        return kernel.astype(np.uint8)

    def create_ellipsoidal_kernel(long_axis, short_axis):
        """Create a 2D ellipsoidal kernel with specified axis lengths and orientation.

        Parameters:
        long_axis (int or float): The length of the long axis.
        short_axis (int or float): The length of the short axis.

        Returns:
        numpy.ndarray: A 2D numpy array representing the ellipsoidal kernel.
        """
        semi_major, semi_minor = long_axis / 2, short_axis / 2

        # Determine the size of the kernel

        size_y = int(np.ceil(semi_minor))
        size_x = int(np.ceil(semi_major))
        
        # Create a grid of (x, y) coordinates centered at (0,0)
        y, x = np.ogrid[-semi_minor:semi_minor+1, -semi_major:semi_major+1]
        
        # Ellipsoid equation: (x/a)^2 + (y/b)^2 <= 1
        ellipse = (x**2 / semi_major**2) + (y**2 / semi_minor**2) <= 1
        
        return ellipse.astype(np.uint8)


    # Function to process each slice
    def process_slice(z):
        tiff_slice = tiff_array[z].astype(np.uint8)
        dilated_slice = cv2.dilate(tiff_slice, kernel, iterations=1)
        return z, dilated_slice

    def process_slice_other(y):
        tiff_slice = tiff_array[:, y, :].astype(np.uint8)
        dilated_slice = cv2.dilate(tiff_slice, kernel, iterations=1)
        return y, dilated_slice

    # Create empty arrays to store the dilated results for the XY and XZ planes
    dilated_xy = np.zeros_like(tiff_array, dtype=np.uint8)
    dilated_xz = np.zeros_like(tiff_array, dtype=np.uint8)

    kernel_x = int(dilated_x)
    kernel = create_circular_kernel(kernel_x)

    num_cores = mp.cpu_count()

    with ThreadPoolExecutor(max_workers=num_cores) as executor:
        futures = {executor.submit(process_slice, z): z for z in range(tiff_array.shape[0])}

        for future in as_completed(futures):
            z, dilated_slice = future.result()
            dilated_xy[z] = dilated_slice

    kernel_x = int(dilated_x)
    kernel_z = int(dilated_z)

    kernel = create_ellipsoidal_kernel(kernel_x, kernel_z)

    with ThreadPoolExecutor(max_workers=num_cores) as executor:
        futures = {executor.submit(process_slice_other, y): y for y in range(tiff_array.shape[1])}
        
        for future in as_completed(futures):
            y, dilated_slice = future.result()
            dilated_xz[:, y, :] = dilated_slice

    # Overlay the results
    final_result = dilated_xy | dilated_xz

    return final_result


def dilate_3D_old(tiff_array, dilated_x, dilated_y, dilated_z):
    """Dilate an array in 3D. Consider replacing with scipy dilation method.
    Arguments are an array,  and the desired pixel dilation amounts in X, Y, Z."""

    # Create empty arrays to store the dilated results for the XY and XZ planes
    dilated_xy = np.zeros_like(tiff_array, dtype=np.uint8)
    dilated_xz = np.zeros_like(tiff_array, dtype=np.uint8)

    # Perform 2D dilation in the XY plane
    for z in range(tiff_array.shape[0]):
        kernel_x = int(dilated_x)
        kernel_y = int(dilated_y)
        kernel = np.ones((kernel_y, kernel_x), dtype=np.uint8)


        # Convert the slice to the appropriate data type
        tiff_slice = tiff_array[z].astype(np.uint8)

        dilated_slice = cv2.dilate(tiff_slice, kernel, iterations=1)
        dilated_xy[z] = dilated_slice

    # Perform 2D dilation in the XZ plane
    for y in range(tiff_array.shape[1]):
        kernel_x = int(dilated_x)
        kernel_z = int(dilated_z)
        kernel = np.ones((kernel_z, kernel_x), dtype=np.uint8)


        # Convert the slice to the appropriate data type
        tiff_slice = tiff_array[:, y, :].astype(np.uint8)

        dilated_slice = cv2.dilate(tiff_slice, kernel, iterations=1)
        dilated_xz[:, y, :] = dilated_slice

    # Overlay the results (you can use logical OR operation or another method)
    final_result = dilated_xy | dilated_xz


    return final_result

def binarize(image):
    """Convert an array from numerical values to boolean mask"""
    return (image != 0).astype(np.uint8)

def invert_array(array):
    """Used to flip glom array indices. 0 becomes 1 and vice versa."""
    return np.logical_not(array).astype(np.uint8)

def process_chunk(start_idx, end_idx, nodes, ring_mask, nearest_label_indices):
    nodes_chunk = nodes[start_idx:end_idx]
    ring_mask_chunk = ring_mask[start_idx:end_idx]
    dilated_nodes_with_labels_chunk = np.copy(nodes_chunk)
    ring_indices = np.argwhere(ring_mask_chunk)

    for index in ring_indices:
        z, y, x = index
        nearest_z, nearest_y, nearest_x = nearest_label_indices[:, z + start_idx, y, x]
        try: #There was an index error here once on the highest val of the second axis. I could not understand why because it usually doesnt hence the try block.
            dilated_nodes_with_labels_chunk[z, y, x] = nodes[nearest_z, nearest_y, nearest_x]
        except:
            pass
    
    return dilated_nodes_with_labels_chunk

def smart_dilate(nodes, dilate_xy, dilate_z, directory = None, GPU = True, fast_dil = False, predownsample = None):

    original_shape = nodes.shape

    # Step 1: Binarize the labeled array
    binary_nodes = binarize(nodes)

    # Step 2: Dilate the binarized array
    if not fast_dil:
        dilated_binary_nodes = dilate_3D(binary_nodes, dilate_xy, dilate_xy, dilate_z)
    else:
        dilated_binary_nodes = dilate_3D_old(binary_nodes, dilate_xy, dilate_xy, dilate_z)


    # Step 3: Isolate the ring (binary dilated mask minus original binary mask)
    ring_mask = dilated_binary_nodes & invert_array(binary_nodes)

    print("Preforming distance transform for smart search... this step may take some time if computed on CPU...")

    try:

        if GPU == True and cp.cuda.runtime.getDeviceCount() > 0:
            print("GPU detected. Using CuPy for distance transform.")

            try:

                if predownsample is None:

                    # Step 4: Find the nearest label for each voxel in the ring
                    nearest_label_indices = compute_distance_transform_GPU(invert_array(nodes))

                else:
                    gotoexcept = 1/0

            except (cp.cuda.memory.OutOfMemoryError, ZeroDivisionError) as e:
                if predownsample is None:
                    down_factor = catch_memory(e) #Obtain downsample amount based on memory missing
                else:
                    down_factor = (predownsample)**3

                while True:
                    downsample_needed = down_factor**(1./3.)
                    small_nodes = nettracer.downsample(nodes, downsample_needed) #Apply downsample
                    try:
                        nearest_label_indices = compute_distance_transform_GPU(invert_array(small_nodes)) #Retry dt on downsample
                        print(f"Using {down_factor} downsample ({downsample_needed} in each dim - Largest possible with this GPU unless user specified downsample)")
                        break
                    except cp.cuda.memory.OutOfMemoryError:
                        down_factor += 1
                binary_nodes = binarize(small_nodes) #Recompute variables for downsample
                dilated_mask = dilated_binary_nodes #Need this for later to stamp out the correct output
                if not fast_dil:
                    dilated_binary_nodes = dilate_3D(binary_nodes, 2 + round_to_odd(dilate_xy/downsample_needed), 2 + round_to_odd(dilate_xy/downsample_needed), 2 + round_to_odd(dilate_z/downsample_needed)) #Mod dilation to recompute variables for downsample while also over dilatiing
                else:
                    dilated_binary_nodes = dilate_3D_old(binary_nodes, 2 + round_to_odd(dilate_xy/downsample_needed), 2 + round_to_odd(dilate_xy/downsample_needed), 2 + round_to_odd(dilate_z/downsample_needed)) 
                ring_mask = dilated_binary_nodes & invert_array(binary_nodes)
                nodes = small_nodes
                del small_nodes
        else:
            goto_except = 1/0
    except Exception as e:
        print("GPU dt failed or did not detect GPU (cupy must be installed with a CUDA toolkit setup...). Computing CPU distance transform instead.")
        if GPU:
            print(f"Error message: {str(e)}")
        nearest_label_indices = compute_distance_transform(invert_array(nodes))


    # Step 5: Process in parallel chunks using ThreadPoolExecutor
    num_cores = mp.cpu_count()  # Use all available CPU cores
    chunk_size = nodes.shape[0] // num_cores  # Divide the array into chunks along the z-axis

    with ThreadPoolExecutor(max_workers=num_cores) as executor:
        args_list = [(i * chunk_size, (i + 1) * chunk_size if i != num_cores - 1 else nodes.shape[0], nodes, ring_mask, nearest_label_indices) for i in range(num_cores)]
        results = list(executor.map(lambda args: process_chunk(*args), args_list))

    # Combine results from chunks
    dilated_nodes_with_labels = np.concatenate(results, axis=0)

    if nodes.shape[1] < original_shape[1]: #If downsample was used, upsample output
        dilated_nodes_with_labels = nettracer.upsample_with_padding(dilated_nodes_with_labels, downsample_needed, original_shape)
        dilated_nodes_with_labels = dilated_nodes_with_labels * dilated_mask

    if directory is not None:
        try:
            tifffile.imwrite(f"{directory}/search_region.tif", dilated_nodes_with_labels)
        except Exception as e:
            print(f"Could not save search region file to {directory}")

    return dilated_nodes_with_labels

def round_to_odd(number):
    rounded = round(number)
    # If the rounded number is even, add or subtract 1 to make it odd
    if rounded % 2 == 0:
        if number > 0:
            rounded += 1
        else:
            rounded -= 1
    return rounded

def smart_label(binary_array, label_array, directory = None, GPU = True, predownsample = None):

    original_shape = binary_array.shape

    if type(binary_array) == str or type(label_array) == str:
        string_bool = True
    else:
        string_bool = None
    if type(binary_array) == str:
        binary_array = tifffile.imread(binary_array)
    if type(label_array) == str:
        label_array = tifffile.imread(label_array)

    # Step 1: Binarize the labeled array
    binary_core = binarize(label_array)
    binary_array = binarize(binary_array)

    # Step 3: Isolate the ring (binary dilated mask minus original binary mask)
    ring_mask = binary_array & invert_array(binary_core)


    try:

        if GPU == True and cp.cuda.runtime.getDeviceCount() > 0:
            print("GPU detected. Using CuPy for distance transform.")

            try:

                if predownsample is None:

                    # Step 4: Find the nearest label for each voxel in the ring
                    nearest_label_indices = compute_distance_transform_GPU(invert_array(binary_core))

                else:
                    gotoexcept = 1/0

            except (cp.cuda.memory.OutOfMemoryError, ZeroDivisionError) as e:
                if predownsample is None:
                    down_factor = catch_memory(e) #Obtain downsample amount based on memory missing
                else:
                    down_factor = (predownsample)**3

                while True:
                    downsample_needed = down_factor**(1./3.)
                    small_array = nettracer.downsample(label_array, downsample_needed) #Apply downsample
                    try:
                        nearest_label_indices = compute_distance_transform_GPU(invert_array(small_array)) #Retry dt on downsample
                        print(f"Using {down_factor} downsample ({downsample_needed} in each dim - Largest possible with this GPU unless user specified downsample)")
                        break
                    except cp.cuda.memory.OutOfMemoryError:
                        down_factor += 1
                binary_mask = binary_array #Need this for later to stamp out the correct output
                binary_core = binarize(small_array)
                label_array = small_array
                binary_array = nettracer.downsample(binary_array, downsample_needed)
                binary_array = nettracer.dilate_3D(binary_array, 3, 3, 3)
                ring_mask = binary_array & invert_array(binary_core)

        else:
            goto_except = 1/0
    except Exception as e:
        if GPU:
            print("GPU dt failed or did not detect GPU (cupy must be installed with a CUDA toolkit setup...). Computing CPU distance transform instead.")
            print(f"Error message: {str(e)}")
            import traceback
            print(traceback.format_exc())
        nearest_label_indices = compute_distance_transform(invert_array(label_array))

    print("Preforming distance transform for smart label...")

    # Step 5: Process in parallel chunks using ThreadPoolExecutor
    num_cores = mp.cpu_count()  # Use all available CPU cores
    chunk_size = label_array.shape[0] // num_cores  # Divide the array into chunks along the z-axis

    with ThreadPoolExecutor(max_workers=num_cores) as executor:
        args_list = [(i * chunk_size, (i + 1) * chunk_size if i != num_cores - 1 else label_array.shape[0], label_array, ring_mask, nearest_label_indices) for i in range(num_cores)]
        results = list(executor.map(lambda args: process_chunk(*args), args_list))

    # Combine results from chunks
    dilated_nodes_with_labels = np.concatenate(results, axis=0)

    if label_array.shape[1] < original_shape[1]: #If downsample was used, upsample output
        dilated_nodes_with_labels = nettracer.upsample_with_padding(dilated_nodes_with_labels, downsample_needed, original_shape)
        dilated_nodes_with_labels = dilated_nodes_with_labels * binary_mask

    if string_bool:
        if directory is not None:
            try:
                tifffile.imwrite(f"{directory}/smart_labelled_array.tif", dilated_nodes_with_labels)
            except Exception as e:
                print(f"Could not save search region file to {directory}")
        else:
            try:
                tifffile.imwrite("smart_labelled_array.tif", dilated_nodes_with_labels)
            except Exception as e:
                print(f"Could not save search region file to active directory")

    return dilated_nodes_with_labels

def compute_distance_transform_GPU(nodes):
    # Convert numpy array to CuPy array
    nodes_cp = cp.asarray(nodes)
    
    # Compute the distance transform on the GPU
    distance, nearest_label_indices = cpx.distance_transform_edt(nodes_cp, return_indices=True)
    
    # Convert results back to numpy arrays
    nearest_label_indices_np = cp.asnumpy(nearest_label_indices)
    
    return nearest_label_indices_np


def compute_distance_transform(nodes):
    distance, nearest_label_indices = distance_transform_edt(nodes, return_indices=True)
    return nearest_label_indices



def compute_distance_transform_distance_GPU(nodes):

    # Convert numpy array to CuPy array
    nodes_cp = cp.asarray(nodes)
    
    # Compute the distance transform on the GPU
    distance, nearest_label_indices = cpx.distance_transform_edt(nodes_cp, return_indices=True)
    
    # Convert results back to numpy arrays
    distance = cp.asnumpy(distance)
    
    return distance    


def compute_distance_transform_distance(nodes):

    # Fallback to CPU if there's an issue with GPU computation
    distance, nearest_label_indices = distance_transform_edt(nodes, return_indices=True)
    return distance




def gaussian(search_region, GPU = True):
    try:
        if GPU == True and cp.cuda.runtime.getDeviceCount() > 0:
            print("GPU detected. Using CuPy for guassian blur.")

            # Convert to CuPy array
            search_region_cp = cp.asarray(search_region)

            # Apply Gaussian filter
            blurred_search_cp = cpx.gaussian_filter(search_region_cp, sigma=1)

            # Convert back to NumPy array if needed
            blurred_search = cp.asnumpy(blurred_search_cp)

            return blurred_search
        else:
            print("Using CPU for guassian blur")
            blurred_search = gaussian_filter(search_region, sigma = 1)
            return blurred_search
    except Exception as e:
        print("GPU blur failed or did not detect GPU (cupy must be installed with a CUDA toolkit setup...). Computing CPU guassian blur instead.")
        print(f"Error message: {str(e)}")
        # Fallback to CPU if there's an issue with GPU computation
        blurred_search = gaussian_filter(search_region, sigma = 1)
        return blurred_search

def get_local_maxima(distance, image):
    try:
        if cp.cuda.runtime.getDeviceCount() > 0:
            print("GPU detected. Using CuPy for local maxima.")

            distance = cp.asarray(distance)

            # Perform a maximum filter to find local maxima
            footprint = cp.ones((3, 3, 3))  # Define your footprint
            filtered = maximum_filter(distance, footprint=footprint)

            # Find local maxima by comparing with the original array
            local_max = (distance == filtered)  # Peaks are where the filtered result matches the original

            # Extract coordinates of local maxima
            coords = cp.argwhere(local_max)
            coords = cp.asnumpy(coords)

            return coords
        else:
            print("No GPU detected. Using CPU for local maxima.")
            coords = peak_local_max(distance, footprint=np.ones((3, 3, 3)), labels=image)
            return coords
    except Exception as e:
        print("GPU operation failed or did not detect GPU (cupy must be installed with a CUDA toolkit set up...). Computing CPU local maxima instead.")
        coords = peak_local_max(distance, footprint=np.ones((3, 3, 3)), labels=image)
        return coords


def catch_memory(e):


    # Get the current GPU device
    device = cp.cuda.Device()

    # Get total memory in bytes
    total_memory = device.mem_info[1]

    # Capture the error message
    error_message = str(e)
    print(f"Error encountered: {error_message}")

    # Use regex to extract the memory required from the error message
    match = re.search(r'allocating ([\d,]+) bytes', error_message)

    if match:
        memory_required = int(match.group(1).replace(',', ''))

        print(f"GPU Memory required for distance transform: {memory_required}, retrying with temporary downsample")

        downsample_needed = (memory_required/total_memory)
        return (downsample_needed)


if __name__ == "__main__":
    nodes = input("Labelled Nodes tiff?: ")
    nodes = tifffile.imread(nodes)

    # Step 1: Binarize the labeled array
    binary_nodes = binarize(nodes)

    # Step 2: Dilate the binarized array
    dilated_binary_nodes = dilate_3D(binary_nodes, 10, 10, 10)

    # Step 3: Isolate the ring (binary dilated mask minus original binary mask)
    ring_mask = dilated_binary_nodes & invert_array(binary_nodes)

    # Step 4: Find the nearest label for each voxel in the ring
    distance, nearest_label_indices = distance_transform_edt(invert_array(nodes), return_indices=True)

    # Step 5: Process in parallel chunks using ThreadPoolExecutor
    num_cores = mp.cpu_count()  # Use all available CPU cores
    chunk_size = nodes.shape[0] // num_cores  # Divide the array into chunks along the z-axis

    with ThreadPoolExecutor(max_workers=num_cores) as executor:
        args_list = [(i * chunk_size, (i + 1) * chunk_size if i != num_cores - 1 else nodes.shape[0], nodes, ring_mask, nearest_label_indices) for i in range(num_cores)]
        results = list(executor.map(lambda args: process_chunk(*args), args_list))

    # Combine results from chunks
    dilated_nodes_with_labels = np.concatenate(results, axis=0)

    # Save the result
    output_file = "dilated_nodes_with_labels.tif"
    tifffile.imwrite(output_file, dilated_nodes_with_labels)
    print(f"Result saved to {output_file}")


from . import nettracer
from . import network_analysis
import numpy as np
from scipy.ndimage import zoom
import multiprocessing as mp
from concurrent.futures import ThreadPoolExecutor, as_completed
import tifffile
import pandas as pd

def get_reslice_indices(args):
    """Internal method used for the secondary algorithm that finds dimensions for subarrays around nodes"""

    indices, dilate_xy, dilate_z, array_shape = args
    try:
        max_indices = np.amax(indices, axis = 0) #Get the max/min of each index.
    except ValueError: #Return Nones if this error is encountered
        return None, None, None
    min_indices = np.amin(indices, axis = 0)

    z_max, y_max, x_max = max_indices[0], max_indices[1], max_indices[2]

    z_min, y_min, x_min = min_indices[0], min_indices[1], min_indices[2]

    y_max = y_max + ((dilate_xy-1)/2) + 1 #Establish dimensions of intended subarray, expanding the max/min indices to include
    y_min = y_min - ((dilate_xy-1)/2) - 1 #the future dilation space (by adding/subtracting half the dilation kernel for each axis)
    x_max = x_max + ((dilate_xy-1)/2) + 1 #an additional index is added in each direction to make sure nothing is discluded.
    x_min = x_min - ((dilate_xy-1)/2) - 1
    z_max = z_max + ((dilate_z-1)/2) + 1
    z_min = z_min - ((dilate_z-1)/2) - 1

    if y_max > (array_shape[1] - 1): #Some if statements to make sure the subarray will not cause an indexerror
        y_max = (array_shape[1] - 1)
    if x_max > (array_shape[2] - 1):
        x_max = (array_shape[2] - 1)
    if z_max > (array_shape[0] - 1):
        z_max = (array_shape[0] - 1)
    if y_min < 0:
        y_min = 0
    if x_min < 0:
        x_min = 0
    if z_min < 0:
        z_min = 0

    y_vals = [y_min, y_max] #Return the subarray dimensions as lists
    x_vals = [x_min, x_max]
    z_vals = [z_min, z_max]

    return z_vals, y_vals, x_vals

def reslice_3d_array(args):
    """Internal method used for the secondary algorithm to reslice subarrays around nodes."""

    input_array, z_range, y_range, x_range = args
    z_start, z_end = z_range
    z_start, z_end = int(z_start), int(z_end)
    y_start, y_end = y_range
    y_start, y_end = int(y_start), int(y_end)
    x_start, x_end = x_range
    x_start, x_end = int(x_start), int(x_end)
    
    # Reslice the array
    resliced_array = input_array[z_start:z_end+1, y_start:y_end+1, x_start:x_end+1]
    
    return resliced_array



def _get_node_edge_dict(label_array, edge_array, label, dilate_xy, dilate_z, cores = 0):
    """Internal method used for the secondary algorithm to find pixel involvement of nodes around an edge."""
    
    # Create a boolean mask where elements with the specified label are True
    label_array = label_array == label
    dil_array = nettracer.dilate_3D(label_array, dilate_xy, dilate_xy, dilate_z) #Dilate the label to see where the dilated label overlaps

    if cores == 0: #For getting the volume of objects. Cores presumes you want the 'core' included in the interaction.
        edge_array = edge_array * dil_array  # Filter the edges by the label in question
        label_array = np.count_nonzero(dil_array)
        edge_array = np.count_nonzero(edge_array) # For getting the interacting skeleton

    elif cores == 1: #Cores being 1 presumes you do not want to 'core' included in the interaction
        label_array = dil_array - label_array
        edge_array = edge_array * label_array
        label_array = np.count_nonzero(label_array)
        edge_array = np.count_nonzero(edge_array) # For getting the interacting skeleton

    elif cores == 2: #Presumes you want skeleton within the core but to only 'count' the stuff around the core for volumes... because of imaging artifacts, perhaps
        edge_array = edge_array * dil_array
        label_array = dil_array - label_array
        label_array = np.count_nonzero(label_array)
        edge_array = np.count_nonzero(edge_array) # For getting the interacting skeleton


    
    args = [edge_array, label_array]

    return args

def process_label(args):
    """Internal method used for the secondary algorithm to process a particular node."""
    nodes, edges, label, dilate_xy, dilate_z, array_shape = args
    print(f"Processing node {label}")
    indices = np.argwhere(nodes == label)
    z_vals, y_vals, x_vals = get_reslice_indices((indices, dilate_xy, dilate_z, array_shape))
    if z_vals is None: #If get_reslice_indices ran into a ValueError, nothing is returned.
        return None, None, None
    sub_nodes = reslice_3d_array((nodes, z_vals, y_vals, x_vals))
    sub_edges = reslice_3d_array((edges, z_vals, y_vals, x_vals))
    return label, sub_nodes, sub_edges


def create_node_dictionary(nodes, edges, num_nodes, dilate_xy, dilate_z, cores = 0):
    """Internal method used for the secondary algorithm to process nodes in parallel."""
    # Initialize the dictionary to be returned
    node_dict = {}

    array_shape = nodes.shape

    # Use ThreadPoolExecutor for parallel execution
    with ThreadPoolExecutor(max_workers=mp.cpu_count()) as executor:
        # First parallel section to process labels
        # List of arguments for each parallel task
        args_list = [(nodes, edges, i, dilate_xy, dilate_z, array_shape) for i in range(1, num_nodes + 1)]

        # Execute parallel tasks to process labels
        results = executor.map(process_label, args_list)

        # Second parallel section to create dictionary entries
        for label, sub_nodes, sub_edges in results:
            executor.submit(create_dict_entry, node_dict, label, sub_nodes, sub_edges, dilate_xy, dilate_z, cores)

    return node_dict

def create_dict_entry(node_dict, label, sub_nodes, sub_edges, dilate_xy, dilate_z, cores = 0):
    """Internal method used for the secondary algorithm to pass around args in parallel."""

    if label is None:
        pass
    else:
        node_dict[label] = _get_node_edge_dict(sub_nodes, sub_edges, label, dilate_xy, dilate_z, cores = cores)


def quantify_edge_node(nodes, edges, search = 0, xy_scale = 1, z_scale = 1, cores = 0, resize = None, save = True, skele = False):

    def save_dubval_dict(dict, index_name, val1name, val2name, filename):

        #index name goes on the left, valname on the right
        df = pd.DataFrame.from_dict(dict, orient='index', columns=[val1name, val2name])

        # Rename the index to 'Node ID'
        df.index.name = index_name

        # Save DataFrame to Excel file
        df.to_excel(filename, engine='openpyxl')

    if type(nodes) is str:
        nodes = tifffile.imread(nodes)

    if type(edges) is str:
        edges = tifffile.imread(edges)

    if skele:
        edges = nettracer.skeletonize(edges)
    else:
        edges = nettracer.binarize(edges)

    if len(np.unique(nodes)) == 2:
        nodes, num_nodes = nettracer.label_objects(nodes)
    else:
        num_nodes = np.max(nodes)

    if resize is not None:
        edges = zoom(edges, resize)
        nodes = zoom(nodes, resize)
        edges = nettracer.skeletonize(edges)

    if search > 0:
        dilate_xy, dilate_z = nettracer.dilation_length_to_pixels(xy_scale, z_scale, search, search)
    else:
        dilate_xy, dilate_z = 0, 0


    edge_quants = create_node_dictionary(nodes, edges, num_nodes, dilate_xy, dilate_z, cores = cores) #Find which edges connect which nodes and put them in a dictionary.

    if save:
    
        save_dubval_dict(edge_quants, 'NodeID', 'Edge Skele Quantity', 'Search Region Volume', 'edge_node_quantity.xlsx')

    else:

        return edge_quants

def calculate_voxel_volumes(array, xy_scale = 1, z_scale = 1):
    """
    Calculate voxel volumes for each uniquely labelled object in a 3D numpy array.
    
    Args:
        array: 3D numpy array where different objects are marked with different integer labels
        
    Returns:
        Dictionary mapping object labels to their voxel volumes
    """
    # Get unique labels (excluding 0 which typically represents background)
    labels = np.unique(array)
    if len(labels) == 2:
        array, _ = nettracer.label_objects(array)
        labels = np.unique(array)

    labels = labels[labels != 0]  # Remove background label if present
    
    # Create dictionary of label: volume pairs
    volumes = {label: (np.count_nonzero(array == label) * (xy_scale**2) * z_scale) for label in labels}
    
    return volumes


import pandas as pd
import networkx as nx
import tifffile
import numpy as np
from networkx.algorithms import community
from community import community_louvain
from scipy.ndimage import zoom
from scipy import ndimage
from . import node_draw
import matplotlib.pyplot as plt
from scipy.optimize import curve_fit
from . import nettracer
from . import modularity
import multiprocessing as mp
import random
from concurrent.futures import ThreadPoolExecutor, as_completed
try:
    import cupy as cp
    import cupyx.scipy.ndimage as cpx
except:
    pass

def downsample(data, factor):
    # Downsample the input data by a specified factor
    return zoom(data, 1/factor, order=0)

def compute_centroid(binary_stack, label):
    """
    Finds centroid of labelled object in array
    """
    indices = np.argwhere(binary_stack == label)
    if indices.shape[0] == 0:
        return None
    else:
        centroid = np.round(np.mean(indices, axis=0)).astype(int)
        
    return centroid

def open_network(excel_file_path):
    """opens an unweighted network from the network excel file"""

    if type(excel_file_path) == str:
        # Read the Excel file into a pandas DataFrame
        master_list = read_excel_to_lists(excel_file_path)
    else:
        master_list = excel_file_path

    # Create a graph
    G = nx.Graph()

    nodes_a = master_list[0]
    nodes_b = master_list[1]

    # Add edges to the graph
    for i in range(len(nodes_a)):
        G.add_edge(nodes_a[i], nodes_b[i])

    return G

def read_excel_to_lists(file_path, sheet_name=0):
    """Convert a pd dataframe to lists. Handles both .xlsx and .csv files"""
    if type(file_path) == str:
        # Check file extension
        if file_path.lower().endswith('.xlsx'):
            # Read the Excel file into a DataFrame without headers
            df = pd.read_excel(file_path, header=None, sheet_name=sheet_name)
            df = df.drop(0)
        elif file_path.lower().endswith('.csv'):
            # Read the CSV file into a DataFrame without headers
            df = pd.read_csv(file_path, header=None)
            df = df.drop(0)
        else:
            raise ValueError("File must be either .xlsx or .csv format")
    else:
        df = file_path

    # Initialize an empty list to store the lists of values
    data_lists = []
    # Iterate over each column in the DataFrame
    for column_name, column_data in df.items():
        # Convert the column values to a list and append to the data_lists
        data_lists.append(column_data.tolist())

    master_list = [[], [], []]
    for i in range(0, len(data_lists), 3):
        master_list[0].extend(data_lists[i])
        master_list[1].extend(data_lists[i+1])
        try:
            master_list[2].extend(data_lists[i+2])
        except IndexError:
            pass
            
    return master_list

def master_list_to_excel(master_list, excel_name):

    nodesA = master_list[0]
    nodesB = master_list[1]
    edgesC = master_list[2]
    # Create a DataFrame from the lists
    df = pd.DataFrame({
    'Nodes A': nodesA,
    'Nodes B': nodesB,
    'Edges C': edgesC
    })

    # Save the DataFrame to an Excel file
    df.to_excel(excel_name, index=False)

def weighted_network(excel_file_path):
    """creates a network where the edges have weights proportional to the number of connections they make between the same structure"""

    if type(excel_file_path) == list:
        master_list = excel_file_path
    else:
        # Read the Excel file into a pandas DataFrame
        master_list = read_excel_to_lists(excel_file_path)

    # Create a graph
    G = nx.Graph()

    # Create a dictionary to store edge weights based on node pairs
    edge_weights = {}

    nodes_a = master_list[0]
    nodes_b = master_list[1]

    # Iterate over the DataFrame rows and update edge weights
    for i in range(len(nodes_a)):
        node1, node2 = nodes_a[i], nodes_b[i]
        edge = (node1, node2) if node1 < node2 else (node2, node1)  # Ensure consistent order
        edge_weights[edge] = edge_weights.get(edge, 0) + 1

    # Add edges to the graph with weights
    for edge, weight in edge_weights.items():
        G.add_edge(edge[0], edge[1], weight=weight)

    return G, edge_weights

def community_partition_simple(nodes, network, centroids = None, down_factor = None, color_code = False, directory = None):

    G = network

    communities = list(nx.community.label_propagation_communities(G))
    partition = {}
    for i, community in enumerate(communities):
        for node in community:
            partition[node] = i + 1
    print(partition)


    print("Generating excel notebook containing community of each node...")
    if type(nodes) == str:
        nodes = tifffile.imread(nodes)

        if len(np.unique(nodes)) < 3:
        
            structure_3d = np.ones((3, 3, 3), dtype=int)
            nodes, num_nodes = ndimage.label(nodes, structure=structure_3d)

    # Convert dictionary to DataFrame with keys as index and values as a column
    df = pd.DataFrame.from_dict(partition, orient='index', columns=['CommunityID'])

    # Rename the index to 'Node ID'
    df.index.name = 'Node ID'

    if directory is None:

        # Save DataFrame to Excel file
        df.to_excel('communities.xlsx', engine='openpyxl')
        print("Community info saved to communities.xls")
    else:
        df.to_excel(f'{directory}/communities.xlsx', engine='openpyxl')
        print(f"Community info saved to {directory}/communities.xls")


    print("Drawing overlay containing community labels for each node...")

    if centroids is None:
        centroids = _find_centroids(nodes, down_factor = down_factor)

    labels = node_draw.degree_draw(partition, centroids, nodes)

    if directory is None:
        tifffile.imwrite("community_labels.tif", labels)
        print("Community labels saved to community_labels.tif")
    else:
        tifffile.imwrite(f"{directory}/community_labels.tif", labels)
        print(f"Community labels saved to {directory}/community_labels.tif")

    print("Drawing overlay containing grayscale community labels for each node...")

    masks = node_draw.degree_infect(partition, nodes)

    if color_code:
        print("And drawing color coded community labels...")
        colored_masks = _color_code(masks)

        if directory is None:

            tifffile.imwrite("community_labels_colorcoded.tif", colored_masks)
            print("Color coded communities saved to community_labels_colorcoded.tif")

        else:
            tifffile.imwrite(f"{directory}/community_labels_colorcoded.tif", colored_masks)
            print(f"Color coded communities saved to {directory}/community_labels_colorcoded.tif")


    if directory is None:

        tifffile.imwrite("community_labels_grayscale.tif", masks)
        print("Grayscale labeled communities saved to community_labels_grayscale.tif")

    else:
        tifffile.imwrite(f"{directory}/community_labels_grayscale.tif", masks)
        print(f"Grayscale labeled communities saved to {directory}/community_labels_grayscale.tif")

    return partition 


def community_partition(nodes, network, centroids = None, down_factor = None, color_code = False, directory = None):

    G, edge_weights = weighted_network(network)

    G = nx.Graph()

    # Find the maximum and minimum edge weights
    max_weight = max(weight for edge, weight in edge_weights.items())
    min_weight = min(weight for edge, weight in edge_weights.items())

    if max_weight > 1:
        # Normalize edge weights to the range [0.1, 1.0]
        normalized_weights = {edge: 0.1 + 0.9 * ((weight - min_weight) / (max_weight - min_weight)) for edge, weight in edge_weights.items()}
    else:
        normalized_weights = {edge: 0.1 for edge, weight in edge_weights.items()}

    # Add edges to the graph with normalized weights
    for edge, normalized_weight in normalized_weights.items():
        G.add_edge(edge[0], edge[1], weight=normalized_weight)

    # Perform Louvain community detection
    partition = community_louvain.best_partition(G)

    for key in partition.keys():
       partition[key] = partition[key] + 1 


    print("Generating excel notebook containing community of each node...")
    if type(nodes) == str:
        nodes = tifffile.imread(nodes)

        if len(np.unique(nodes)) < 3:
        
            structure_3d = np.ones((3, 3, 3), dtype=int)
            nodes, num_nodes = ndimage.label(nodes, structure=structure_3d)

    # Convert dictionary to DataFrame with keys as index and values as a column
    df = pd.DataFrame.from_dict(partition, orient='index', columns=['CommunityID'])

    # Rename the index to 'Node ID'
    df.index.name = 'Node ID'

    if directory is None:

        # Save DataFrame to Excel file
        df.to_excel('communities.xlsx', engine='openpyxl')
        print("Community info saved to communities.xls")
    else:
        df.to_excel(f'{directory}/communities.xlsx', engine='openpyxl')
        print(f"Community info saved to {directory}/communities.xls")


    print("Drawing overlay containing community labels for each node...")

    if centroids is None:
        centroids = _find_centroids(nodes, down_factor = down_factor)

    labels = node_draw.degree_draw(partition, centroids, nodes)

    if directory is None:
        tifffile.imwrite("community_labels.tif", labels)
        print("Community labels saved to community_labels.tif")
    else:
        tifffile.imwrite(f"{directory}/community_labels.tif", labels)
        print(f"Community labels saved to {directory}/community_labels.tif")

    print("Drawing overlay containing grayscale community labels for each node...")

    masks = node_draw.degree_infect(partition, nodes)

    if color_code:
        print("And drawing color coded community labels...")
        colored_masks = _color_code(masks)

        if directory is None:

            tifffile.imwrite("community_labels_colorcoded.tif", colored_masks)
            print("Color coded communities saved to community_labels_colorcoded.tif")

        else:
            tifffile.imwrite(f"{directory}/community_labels_colorcoded.tif", colored_masks)
            print(f"Color coded communities saved to {directory}/community_labels_colorcoded.tif")


    if directory is None:

        tifffile.imwrite("community_labels_grayscale.tif", masks)
        print("Grayscale labeled communities saved to community_labels_grayscale.tif")

    else:
        tifffile.imwrite(f"{directory}/community_labels_grayscale.tif", masks)
        print(f"Grayscale labeled communities saved to {directory}/community_labels_grayscale.tif")

    return partition



def _color_code(grayscale_image):
    """Color code a grayscale array. Currently expects linearly ascending grayscale labels, will crash if there are gaps. (Main use case is grayscale anyway)"""

    def generate_colormap(num_labels):
        # Generate a colormap with visually distinct colors using the new method
        cmap = plt.colormaps['hsv']
        colors = cmap(np.linspace(0, 1, num_labels))
        return colors

    def grayscale_to_rgb(grayscale_image):
        # Get the number of labels
        num_labels = np.max(grayscale_image) + 1
        
        # Generate a colormap
        colormap = generate_colormap(num_labels)
        
        # Create an empty RGB image
        rgb_image = np.zeros((*grayscale_image.shape, 3), dtype=np.uint8)
        
        # Assign colors to each label
        for label in range(1, num_labels):
            color = (colormap[label][:3] * 255).astype(np.uint8)  # Convert to RGB and ensure dtype is uint8
            rgb_image[grayscale_image == label] = color
            
        return rgb_image

    # Convert the grayscale image to RGB
    rgb_image = grayscale_to_rgb(grayscale_image)

    return rgb_image

def read_centroids_to_dict(file_path):
    """
    Read centroid data from either Excel (.xlsx) or CSV file into a dictionary.
    
    Parameters:
    file_path (str): Path to the input file (.xlsx or .csv)
    
    Returns:
    dict: Dictionary with first column as keys and next three columns as numpy array values
    """
    # Check file extension
    if file_path.lower().endswith('.xlsx'):
        df = pd.read_excel(file_path)
    elif file_path.lower().endswith('.csv'):
        df = pd.read_csv(file_path)
    else:
        raise ValueError("Unsupported file format. Please provide either .xlsx or .csv file")
    
    # Initialize an empty dictionary
    data_dict = {}
    
    # Iterate over each row in the DataFrame
    for _, row in df.iterrows():
        # First column is the key
        key = row.iloc[0]
        # Next three columns are the values
        value = np.array(row.iloc[1:4])
        # Add the key-value pair to the dictionary
        data_dict[key] = value
        
    return data_dict

def read_excel_to_singval_dict(file_path):
    """
    Read data from either Excel (.xlsx) or CSV file into a dictionary with single values.
    
    Parameters:
    file_path (str): Path to the input file (.xlsx or .csv)
    
    Returns:
    dict: Dictionary with first column as keys and second column as values
    """
    # Check file extension and read accordingly
    if file_path.lower().endswith('.xlsx'):
        df = pd.read_excel(file_path)
    elif file_path.lower().endswith('.csv'):
        df = pd.read_csv(file_path)
    else:
        raise ValueError("Unsupported file format. Please provide either .xlsx or .csv file")
    
    # Convert the DataFrame to a dictionary
    data_dict = {}
    for idx, row in df.iterrows():
        key = row.iloc[0]  # First column as key
        value = row.iloc[1]  # Second column as value
        data_dict[key] = value
        
    return data_dict

def combine_lists_to_sublists(master_list):

    def fill_if_empty(lst, num_zeros):
        """
        Checks if a list is empty and fills it with zeros if it is.
        
        Args:
            lst (list): The list to check and potentially fill
            num_zeros (int): Number of zeros to add if list is empty
        
        Returns:
            list: The original list if not empty, or new list filled with zeros if empty
        """
        if not lst:  # This checks if the list is empty
            lst.extend([0] * num_zeros)
        return lst

    list1 = master_list[0]
    list2 = master_list[1]
    list3 = master_list[2]
    list3 = fill_if_empty(list3, len(list1))

    # Combine the lists into one list of sublists
    combined_list = [list(sublist) for sublist in zip(list1, list2, list3)]
    
    return combined_list


def find_centroids(nodes, down_factor = None, network = None):

    """Can be used to save an excel file containing node IDs and centroids in a network. Inputs are a node.tif or node np array, an optional network excel file, and optional downsample factor"""

    if type(nodes) == str: #Open into numpy array if filepath
        nodes = tifffile.imread(nodes)

    if len(np.unique(nodes)) == 2: #Label if binary
        structure_3d = np.ones((3, 3, 3), dtype=int)
        nodes, num_nodes = ndimage.label(nodes)

    if down_factor is not None:
        nodes = downsample(nodes, down_factor)
    else: 
        down_factor = 1

    centroid_dict = {}

    if network is not None:

        G = open_network(network)

        node_ids = list(G.nodes)

        for nodeid in node_ids:
            centroid = compute_centroid(nodes, nodeid)
            if centroid is not None:
                centroid = down_factor * centroid
                centroid_dict[nodeid] = centroid

    else:
        node_max = np.max(nodes)
        for nodeid in range(1, node_max + 1):
            centroid = compute_centroid(nodes, nodeid)
            if centroid is not None:
                centroid = down_factor * centroid
                centroid_dict[nodeid] = centroid

    _save_centroid_dictionary(centroid_dict)

    return centroid_dict

def _save_centroid_dictionary(centroid_dict, filepath = None, index = 'Node ID'):
    # Convert dictionary to DataFrame with keys as index and values as a column
    #for item in centroid_dict:
        #representative = centroid_dict[item]
        #break

    #if len(representative) == 3:
        #df = pd.DataFrame.from_dict(centroid_dict, orient='index', columns=['Z', 'Y', 'X'])
    #elif len(representative) == 2:
        #df = pd.DataFrame.from_dict(centroid_dict, orient='index', columns=['Y', 'X'])

    df = pd.DataFrame.from_dict(centroid_dict, orient='index', columns=['Z', 'Y', 'X'])

    # Rename the index to 'Node ID'
    df.index.name = index

    if filepath is None:
        try:
            # Save DataFrame to Excel file
            df.to_excel('centroids.xlsx', engine='openpyxl')
        except Exception as e:
            print("Could not save centroids to active directory")
    else:
        try:
            # Save DataFrame to Excel file
            df.to_excel(filepath, engine='openpyxl')
        except Exception as e:
            print(f"Could not save centroids to {filepath}")

def _find_centroids_GPU(nodes, node_list=None, down_factor=None):
    """Internal use version to get centroids without saving"""

    def _compute_centroid_GPU(binary_stack, label):
        """
        Finds centroid of labelled object in array
        """
        indices = cp.argwhere(binary_stack == label)
        if indices.shape[0] == 0:
            return None
        else:
            centroid = cp.round(np.mean(indices, axis=0)).astype(int)
    
        centroid = centroid.tolist()
        return centroid

    nodes = cp.asarray(nodes)
    if isinstance(nodes, str):  # Open into numpy array if filepath
        nodes = tifffile.imread(nodes)

        if len(cp.unique(nodes)) == 2:  # Label if binary
            structure_3d = cp.ones((3, 3, 3), dtype=int)
            nodes, num_nodes = cpx.label(nodes)

    if down_factor is not None:
        nodes = cp.asnumpy(nodes)
        nodes = downsample(nodes, down_factor)
        nodes = cp.asarray(nodes)
    else:
        down_factor = 1

    centroid_dict = {}

    if node_list is None:
        node_list = cp.unique(nodes)
        node_list = node_list.tolist()
        if node_list[0] == 0:
            del node_list[0]

    for label in node_list:
        centroid = _compute_centroid_GPU(nodes, label)
        if centroid is not None:
            centroid_dict[label] = centroid

    return centroid_dict

def _find_centroids_old(nodes, node_list = None, down_factor = None):

    """Internal use version to get centroids without saving"""


    if type(nodes) == str: #Open into numpy array if filepath
        nodes = tifffile.imread(nodes)

        if len(np.unique(nodes)) == 2: #Label if binary
            structure_3d = np.ones((3, 3, 3), dtype=int)
            nodes, num_nodes = ndimage.label(nodes)

    if down_factor is not None:
        nodes = downsample(nodes, down_factor)
    else:
        down_factor = 1

    centroid_dict = {}

    if node_list is None:

        node_max = np.max(nodes)

        for nodeid in range(1, node_max + 1):
            centroid = compute_centroid(nodes, nodeid)
            if centroid is not None:
                #centroid = down_factor * centroid
                centroid_dict[nodeid] = centroid

    else:
        for nodeid in node_list:
            centroid = compute_centroid(nodes, nodeid)
            if centroid is not None:
                #centroid = down_factor * centroid
                centroid_dict[nodeid] = centroid

    return centroid_dict

def _find_centroids(nodes, node_list=None, down_factor=None):
    """Internal use version to get centroids without saving"""

    def _compute_centroid(binary_stack, label, z_offset):
        """
        Finds centroid of labelled object in array and adjusts for the Z-offset.
        """
        indices = np.argwhere(binary_stack == label)
        centroid = np.round(np.mean(indices, axis=0)).astype(int)
        # Adjust the Z coordinate by the z_offset
        centroid[0] += z_offset
        return centroid

    def compute_centroids_in_chunk(chunk, z_offset):
        """
        Compute centroids for all labels in a given chunk of the 3D array.
        Adjust Z-coordinate based on the z_offset for each chunk.
        """
        centroid_dict_chunk = {}
        label_list = np.unique(chunk)
        try:
            if label_list[0] == 0:
                np.delete(label_list, 0)
        except:
            pass

        for label in label_list:
            centroid = _compute_centroid(chunk, label, z_offset)
            centroid_dict_chunk[label] = centroid
        return centroid_dict_chunk

    def chunk_3d_array(array, num_chunks):
        """
        Split the 3D array into smaller chunks along the z-axis.
        """
        z_slices = np.array_split(array, num_chunks, axis=0)
        return z_slices

    if isinstance(nodes, str):  # Open into numpy array if filepath
        nodes = tifffile.imread(nodes)

        if len(np.unique(nodes)) == 2:  # Label if binary
            structure_3d = np.ones((3, 3, 3), dtype=int)
            nodes, num_nodes = ndimage.label(nodes)

    if down_factor is not None:
        nodes = downsample(nodes, down_factor)
    else:
        down_factor = 1

    centroid_dict = {}

    num_cpus = mp.cpu_count()
    # Chunk the 3D array along the z-axis into smaller subarrays
    node_chunks = chunk_3d_array(nodes, num_cpus)

    # Calculate Z offset for each chunk
    chunk_sizes = [chunk.shape[0] for chunk in node_chunks]
    z_offsets = np.cumsum([0] + chunk_sizes[:-1])

    # Parallel computation of centroids across chunks
    with ThreadPoolExecutor(max_workers=num_cpus) as executor:
        futures = {executor.submit(compute_centroids_in_chunk, chunk, z_offset): chunk_id
                   for chunk_id, (chunk, z_offset) in enumerate(zip(node_chunks, z_offsets))}
        
        for future in as_completed(futures):
            centroids_chunk = future.result()
            centroid_dict.update(centroids_chunk)

    try:
        del centroid_dict[0]
    except:
        pass

    return centroid_dict


def get_degrees(nodes, network, down_factor = None, directory = None, centroids = None, called = False, no_img = 0):

    print("Generating table containing degree of each node...")
    if type(nodes) == str:
        nodes = tifffile.imread(nodes)

    if len(np.unique(nodes)) < 3:
        
        structure_3d = np.ones((3, 3, 3), dtype=int)
        nodes, num_nodes = ndimage.label(nodes, structure=structure_3d)

    if type(network) == str:

        G, weights = weighted_network(network)
    else:
        G = network

    node_list = list(G.nodes)
    node_dict = {}

    for node in node_list:
        node_dict[node] = (G.degree(node))

    if not called:

        # Convert dictionary to DataFrame with keys as index and values as a column
        df = pd.DataFrame.from_dict(node_dict, orient='index', columns=['Degree'])

        # Rename the index to 'Node ID'
        df.index.name = 'Node ID'

    if not called:

        if directory is None:
            # Save DataFrame to Excel file
            df.to_excel('degrees.xlsx', engine='openpyxl')
            print("Degrees saved to degrees.xlsx")

        else:
            df.to_excel(f'{directory}/degrees.xlsx', engine='openpyxl')
            print(f"Degrees saved to {directory}/degrees.xlsx")


    print("Drawing overlay containing degree labels for each node...")

    if down_factor is not None:

        for item in nodes.shape:
            if item < 5:
                down_factor = 1
                break


    if no_img == 1:

        if centroids is None:

            centroids = _find_centroids(nodes, down_factor = down_factor)

        nodes = node_draw.degree_draw(node_dict, centroids, nodes)

        if not called:

            if directory is None:

                tifffile.imwrite("degree_labels.tif", labels)
                print(f"Degree labels saved to degree_labels.tif")


            else:
                tifffile.imwrite(f"{directory}/degree_labels.tif", labels)
                print(f"Degree labels saved to {directory}/degree_labels.tif")


    elif no_img == 2:

        print("Drawing overlay containing grayscale degree labels for each node...")

        nodes = node_draw.degree_infect(node_dict, nodes)

        if not called:

            if directory is None:

                tifffile.imwrite("degree_labels_grayscale.tif", masks)

            else:
                tifffile.imwrite(f"{directory}/degree_labels_grayscale.tif", masks)

    return node_dict, nodes



def remove_dupes(network):
    """Removes duplicate node connections from network"""
    if type(network) == str:
        network = read_excel_to_lists

    compare_list = []

    nodesA = network[0]
    nodesB = network[1]
    edgesC = network[2]

    # Iterate in reverse order to safely delete elements
    for i in range(len(nodesA) - 1, -1, -1):
        item = [nodesA[i], nodesB[i]]
        if item in compare_list:
            del nodesA[i]
            del nodesB[i]
            del edgesC[i]
            continue
        else:
            compare_list.append(item)

    master_list = [nodesA, nodesB, edgesC]

    return master_list







#Concerning radial analysis:
def radial_analysis(nodes, network, rad_dist, xy_scale = None, z_scale = None, centroids = None, directory = None, down_factor = None):
    print("Performing Radial Distribution Analysis...")

    print("Generating excel notebook containing degree of each node...")
    if type(nodes) == str:
        nodes = tifffile.imread(nodes)

    if len(np.unique(nodes)) < 3:
        
        structure_3d = np.ones((3, 3, 3), dtype=int)
        nodes, num_nodes = ndimage.label(nodes, structure=structure_3d)

    if type(network) == str:

        network = read_excel_to_lists(network)

    if xy_scale is None:
        xy_scale = 1

    if z_scale is None:
        z_scale = 1

    if down_factor is not None:
        xy_scale = xy_scale * down_factor
        z_scale = z_scale * down_factor
        nodes = downsample(nodes, down_factor)


    num_objects = np.max(nodes)

    if centroids is None:
        centroids = _find_centroids(nodes)

    dist_list = get_distance_list(centroids, network, xy_scale, z_scale)
    x_vals, y_vals = buckets(dist_list, num_objects, rad_dist, directory = directory)
    histogram(x_vals, y_vals, directory = directory)
    return dist_list

def buckets(dists, num_objects, rad_dist, directory = None):
    y_vals = []
    x_vals = []
    radius = 0
    max_dist = max(dists)

    while radius < max_dist:
        radius2 = radius + rad_dist
        radial_objs = 0
        for item in dists:
            if item >= radius and item <= radius2:
                radial_objs += 1
        radial_avg = radial_objs/num_objects
        radius = radius2
        x_vals.append(radial_avg)
        y_vals.append(radius)

    # Create a DataFrame from the lists
    data = {'Radial Distance From Any Node': y_vals, 'Average Number of Neighboring Nodes': x_vals}
    df = pd.DataFrame(data)

    if directory is None:
        # Save the DataFrame to an Excel file
        df.to_excel('radial_distribution.xlsx', index=False)
        print("Radial distribution saved to radial_distribution.xlsx")
    else:
        df.to_excel(f'{directory}/radial_distribution.xlsx', index=False)
        print(f"Radial distribution saved to {directory}/radial_distribution.xlsx")

    return x_vals, y_vals

def histogram(counts, y_vals, directory = None):
    # Calculate the bin edges based on the y_vals
    bins = np.linspace(min(y_vals), max(y_vals), len(y_vals) + 1)

    # Create a histogram
    plt.hist(x=y_vals, bins=bins, weights=counts, edgecolor='black')

    # Adding labels and title (Optional, but recommended for clarity)
    plt.title('Radial Distribution of Network')
    plt.xlabel('Distance from any node')
    plt.ylabel('Avg Number of Neigbhoring Vertices')

    if directory is not None:
        plt.savefig(f'{directory}/radial_plot.png')

    # Show the plot
    plt.show()

def get_distance_list(centroids, network, xy_scale, z_scale):
    print("Generating radial distribution...")

    distance_list = [] #init empty list to contain all distance vals

    nodesa = network[0]
    nodesb = network[1]

    for i in range(len(nodesa)):
        try:
            z1, y1, x1 = centroids[nodesa[i]]
            z1, y1, x1 = z1 * z_scale, y1 * xy_scale, x1 * xy_scale
            z2, y2, x2 = centroids[nodesb[i]]
            z2, y2, x2 = z2 * z_scale, y2 * xy_scale, x2 * xy_scale

            dist = np.sqrt((z2 - z1)**2 + (y2 - y1)**2 + (x2 - x1)**2)
            distance_list.append(dist)
        except:
            pass

    return distance_list


def prune_samenode_connections(networkfile, nodeIDs):
    """Remove all connections between nodes of the same ID, to evaluate solely connections to other objects"""

    if type(nodeIDs) == str:
        # Read the Excel file into a DataFrame
        df = pd.read_excel(nodeIDs)
    
        # Convert the DataFrame to a dictionary
        data_dict = pd.Series(df.iloc[:, 1].values, index=df.iloc[:, 0]).to_dict()
    else:
        data_dict = nodeIDs

    if type(networkfile) == str:
        # Read the network file into lists
        master_list = read_excel_to_lists(networkfile)
    else:
        master_list = networkfile

    nodesA = master_list[0]
    nodesB = master_list[1]
    edgesC = master_list[2]

    # Iterate in reverse order to safely delete elements
    for i in range(len(nodesA) - 1, -1, -1):
        nodeA = nodesA[i]
        nodeB = nodesB[i]

        if data_dict.get(nodeA) == data_dict.get(nodeB):
            # Remove the item from all lists
            del nodesA[i]
            del nodesB[i]
            try:
                del edgesC[i]
            except:
                pass

    save_list = []

    for i in range(len(nodesA)):
        try:
            item = [nodesA[i], nodesB[i], edgesC[i]]
        except:
            item = [nodesA[i], nodesB[i], None]
        save_list.append(item)
        
    if type(networkfile) == str:

        filename = 'network_pruned_away_samenode_connections.xlsx'

        nettracer.create_and_save_dataframe(save_list, filename)

        print(f"Pruned network saved to {filename}")

    output_dict = data_dict.copy()
    for item in data_dict:
        if item not in nodesA and item not in nodesB:
            del output_dict[item]

    filename = 'Node_identities_pruned_away_samenode_connections.xlsx'

    if type(networkfile) == str:

        save_singval_dict(output_dict, 'NodeID', 'Identity', filename)

        print(f"Pruned network identities saved to {filename}")

    master_list = [nodesA, nodesB, edgesC]


    # Optional: Return the updated lists if needed
    return master_list, output_dict


def isolate_internode_connections(networkfile, nodeIDs, ID1, ID2):
    """Isolate only connections between two specific node identified elements of a network"""

    if type(nodeIDs) == str:
        """Remove all connections between nodes of the same ID, to evaluate solely connections to other objects"""
        # Read the Excel file into a DataFrame
        df = pd.read_excel(nodeIDs)
        
        # Convert the DataFrame to a dictionary
        data_dict = pd.Series(df.iloc[:, 1].values, index=df.iloc[:, 0]).to_dict()
    else:
        data_dict = nodeIDs

    if type(networkfile) == str:
        # Read the network file into lists
        master_list = read_excel_to_lists(networkfile)
    else:
        master_list = networkfile

    nodesA = master_list[0]
    nodesB = master_list[1]
    edgesC = master_list[2]

    legalIDs = [ID1, ID2]

    for i in range(len(nodesA) - 1, -1, -1):
        nodeA = nodesA[i]
        nodeB = nodesB[i]
        
        valueA = str(data_dict.get(nodeA))
        valueB = str(data_dict.get(nodeB))

        # Check if both values are not in legalIDs
        if valueA not in legalIDs or valueB not in legalIDs:
            # Remove the item from all lists
            del nodesA[i]
            del nodesB[i]
            del edgesC[i]

    save_list = []

    for i in range(len(nodesA)):
        item = [nodesA[i], nodesB[i], edgesC[i]]
        save_list.append(item)


    if type(networkfile) == str:
        filename = f'network_isolated_{ID1}_{ID2}_connections.xlsx'

        nettracer.create_and_save_dataframe(save_list, filename)

        print(f"Isolated internode network saved to {filename}")

    output_dict = data_dict.copy()
    for item in data_dict:
        if item not in nodesA and item not in nodesB:
            del output_dict[item]

    if type(networkfile) == str:

        filename = f'Node_identities_for_isolated_{ID1}_{ID2}_network.xlsx'

        save_singval_dict(output_dict, 'NodeID', 'Identity', filename)

        print(f"Isolated network identities saved to {filename}")

    master_list = [nodesA, nodesB, edgesC]

    # Optional: Return the updated lists if needed
    return master_list, output_dict

def edge_to_node(network, node_identities = None):
    """Converts edge IDs into nodes, so that the node-edge relationships can be more easily visualized"""

    if node_identities is not None and type(node_identities) == str:
        # Read the Excel file into a DataFrame
        df = pd.read_excel(node_identities)
        
        # Convert the DataFrame to a dictionary
        identity_dict = pd.Series(df.iloc[:, 1].values, index=df.iloc[:, 0]).to_dict()
    elif node_identities is not None and type(node_identities) != str:
        identity_dict = node_identities
    else:
        identity_dict = {}

    new_network = []

    # Read the network file into lists
    if type(network) == str:
        master_list = read_excel_to_lists(network)
    else:
        master_list = network

    nodesA = master_list[0]
    nodesB = master_list[1]
    edgesC = master_list[2]
    allnodes = set(nodesA + nodesB)
    maxnode = max(allnodes)
    print(f"Transposing all edge vals by {maxnode} to prevent ID overlap with preexisting nodes")


    for i in range(len(edgesC)):
        edgesC[i] = edgesC[i] + maxnode

    alledges = set(edgesC)

    for i in range(len(edgesC)):
        newpair1 = [nodesA[i], edgesC[i], None]
        newpair2 = [edgesC[i], nodesB[i], None]
        new_network.append(newpair1)
        new_network.append(newpair2)

    for item in allnodes:
        if item not in identity_dict:
            identity_dict[item] = 'Node'

    for item in alledges:
        identity_dict[item] = 'Edge'

    if type(network) == str:

        save_singval_dict(identity_dict, 'NodeID', 'Identity', 'edge_to_node_identities.xlsx')

        nettracer.create_and_save_dataframe(new_network, 'edge-node_network.xlsx')

    else:

        df = nettracer.create_and_save_dataframe(new_network)
        return df, identity_dict, maxnode


def save_singval_dict(dict, index_name, valname, filename):

    #index name goes on the left, valname on the right
    df = pd.DataFrame.from_dict(dict, orient='index', columns=[valname])

    # Rename the index to 'Node ID'
    df.index.name = index_name

    # Save DataFrame to Excel file
    df.to_excel(filename, engine='openpyxl')


def rand_net_weighted(num_rows, num_nodes, nodes):

    random_network = []
    k = 0

    while k < num_rows:

        for i in range(0, num_nodes):
            random_partner = random.randint(0, len(nodes)-1)
            random_partner = nodes[random_partner]
            if random_partner == nodes[i]:
                while random_partner == nodes[i]:
                    random_partner = random.randint(0, len(nodes)-1)
                    random_partner = nodes[random_partner]
            random_pair = [nodes[i], random_partner, None]
            random_network.append(random_pair)
            k+= 1
            if k == num_rows:
                break
        random.shuffle(nodes)


    df = nettracer.create_and_save_dataframe(random_network)

    G, edge_weights = weighted_network(df)

    return G

def rand_net(num_rows, num_nodes, nodes):
    random_network = []
    k=0

    while k < num_rows:
        for i in range(num_nodes):
            # Generate a new random partner until it's valid
            while True:
                random_partner_index = random.randint(0, len(nodes) - 1)
                random_partner = nodes[random_partner_index]
                
                # Check if the random partner is different from the current node
                # and if the pair is not already in the network
                if random_partner != nodes[i] and [nodes[i], random_partner, None] not in random_network and [random_partner, nodes[i], None] not in random_network:
                    break

            random_pair = [nodes[i], random_partner, None]
            random_network.append(random_pair)
            k += 1

            if k == num_rows:
                break


        # Shuffle nodes for the next iteration
        random.shuffle(nodes)



    df = nettracer.create_and_save_dataframe(random_network)

    G, edge_weights = weighted_network(df)

    return G

def generate_random(G, net_lists, weighted = True):

    nodes = list(G.nodes)

    num_nodes = len(nodes)

    num_rows = len(net_lists[0])


    if weighted:

        G = rand_net_weighted(num_rows, num_nodes, nodes)

    else:

        G = rand_net(num_rows, num_nodes, nodes)

    return G

def list_trim(list1, list2, component):

    list1_copy = list1
    indices_to_delete = []
    for i in range(len(list1)):

        if list1[i] not in component and list2[i] not in component:
            indices_to_delete.append(i)

    for i in reversed(indices_to_delete):
        del list1_copy[i]

    return list1_copy

def degree_distribution(G, directory = None):

    def create_incremental_list(length, start=1):
        return list(range(start, start + length))

    node_list = list(G.nodes)
    degree_dict = {}

    for node in node_list:
        degree = G.degree(node)
        if degree not in degree_dict:
            degree_dict[degree] = 1
        else:
            degree_dict[degree] += 1

    high_degree = max(degree_dict.keys())
    proportion_list = [0] * high_degree

    for item in degree_dict:
        proportion_list[item - 1] = float(degree_dict[item]/len(node_list))
    degrees = create_incremental_list(high_degree)


    df = pd.DataFrame({
        'Degree (k)': degrees,
        'Proportion of nodes with degree (p(k))': proportion_list
    })

    if directory is None:
        # Save the DataFrame to an Excel file
        df.to_excel('degree_dist.xlsx', index=False)
        print("Degree distribution saved to degree_dist.xlsx")
    else:
        df.to_excel(f'{directory}/degree_dist.xlsx', index=False)
        print(f"Degree distribution saved to {directory}/degree_dist.xlsx")


    power_trendline(degrees, proportion_list, directory = directory)

    return_dict = {}
    for i in range(len(degrees)):
        return_dict[degrees[i]] = proportion_list[i]

    return return_dict


def power_trendline(x, y, directory = None):
    # Handle zeros in y for logarithmic transformations
    y = np.array(y)
    x = np.array(x)
    y[y == 0] += 0.001

    # Define the power function
    def power_func(x, a, b):
        return a * (x ** b)

    # Fit the power function to the data
    popt, pcov = curve_fit(power_func, x, y)
    a, b = popt

    # Create a range of x values for the trendline
    x_fit = np.linspace(min(x), max(x), 100)
    y_fit = power_func(x_fit, a, b)

    # Calculate R-squared value
    y_pred = power_func(x, *popt)
    ss_res = np.sum((y - y_pred) ** 2)
    ss_tot = np.sum((y - np.mean(y)) ** 2)
    r2 = 1 - (ss_res / ss_tot)

    # Create a scatterplot
    plt.scatter(x, y, label='Data')
    plt.xlabel('Degree (k)')
    plt.ylabel('Proportion of nodes with degree (p(k))')
    plt.title('Degree Distribution of Network')

    # Plot the power trendline
    plt.plot(x_fit, y_fit, color='red', label=f'Power Trendline: $y = {a:.2f}x^{{{b:.2f}}}$')

    # Annotate the plot with the trendline equation and R-squared value
    plt.text(
        0.05, 0.95, 
        f'$y = {a:.2f}x^{{{b:.2f}}}$\n$R^2 = {r2:.2f}$',
        transform=plt.gca().transAxes,
        fontsize=12,
        verticalalignment='top'
    )

    if directory is not None:
        plt.savefig(f'{directory}/degree_plot.png')

    plt.show()




import pandas as pd
import networkx as nx
import tifffile
import numpy as np
from networkx.algorithms import community
from scipy import ndimage
from scipy.ndimage import zoom
from networkx.algorithms import community
from community import community_louvain
from . import node_draw


def binarize(image):
    """Convert an array from numerical values to boolean mask"""
    image = image != 0

    image = image.astype(np.uint8)

    return image

def upsample_with_padding(data, factor, original_shape):
    # Upsample the input binary array while adding padding to match the original shape

    # Get the dimensions of the original and upsampled arrays
    original_shape = np.array(original_shape)
    binary_array = zoom(data, factor, order=0)
    upsampled_shape = np.array(binary_array.shape)

    # Calculate the positive differences in dimensions
    difference_dims = original_shape - upsampled_shape

    # Calculate the padding amounts for each dimension
    padding_dims = np.maximum(difference_dims, 0)
    padding_before = padding_dims // 2
    padding_after = padding_dims - padding_before

    # Pad the binary array along each dimension
    padded_array = np.pad(binary_array, [(padding_before[0], padding_after[0]),
                                         (padding_before[1], padding_after[1]),
                                         (padding_before[2], padding_after[2])], mode='constant', constant_values=0)

    # Calculate the subtraction amounts for each dimension
    sub_dims = np.maximum(-difference_dims, 0)
    sub_before = sub_dims // 2
    sub_after = sub_dims - sub_before

    # Remove planes from the beginning and end
    if sub_dims[0] == 0:
        trimmed_planes = padded_array
    else:
        trimmed_planes = padded_array[sub_before[0]:-sub_after[0], :, :]

    # Remove rows from the beginning and end
    if sub_dims[1] == 0:
        trimmed_rows = trimmed_planes
    else:
        trimmed_rows = trimmed_planes[:, sub_before[1]:-sub_after[1], :]

    # Remove columns from the beginning and end
    if sub_dims[2] == 0:
        trimmed_array = trimmed_rows
    else:
        trimmed_array = trimmed_rows[:, :, sub_before[2]:-sub_after[2]]

    return trimmed_array

def weighted_network(excel_file_path):
    """creates a network where the edges have weights proportional to the number of connections they make between the same structure"""
    # Read the Excel file into a pandas DataFrame
    master_list = read_excel_to_lists(excel_file_path)

    # Create a graph
    G = nx.Graph()

    # Create a dictionary to store edge weights based on node pairs
    edge_weights = {}

    nodes_a = master_list[0]
    nodes_b = master_list[1]

    # Iterate over the DataFrame rows and update edge weights
    for i in range(len(nodes_a)):
        node1, node2 = nodes_a[i], nodes_b[i]
        edge = (node1, node2) if node1 < node2 else (node2, node1)  # Ensure consistent order
        edge_weights[edge] = edge_weights.get(edge, 0) + 1

    # Add edges to the graph with weights
    for edge, weight in edge_weights.items():
        G.add_edge(edge[0], edge[1], weight=weight)

    return G, edge_weights

def compute_centroid(binary_stack, label):
    """
    Finds centroid of labelled object in array
    """
    indices = np.argwhere(binary_stack == label)
    centroid = np.round(np.mean(indices, axis=0)).astype(int)

    return centroid



def get_border_nodes(partition, G):
# Find nodes that border nodes in other communities
    border_nodes = set()
    intercom_connections = 0
    connected_coms = []
    for edge in G.edges():
        if partition[edge[0]] != partition[edge[1]]:
            border_nodes.add(edge[0])
            border_nodes.add(edge[1])
            connected_coms.append(partition[edge[0]])
            connected_coms.append(partition[edge[1]])
            intercom_connections += 1

    return border_nodes, intercom_connections, set(connected_coms)

def downsample(data, factor):
    # Downsample the input data by a specified factor
    return zoom(data, 1/factor, order=0)

def labels_to_boolean(label_array, labels_list):
    # Use np.isin to create a boolean array with a single operation
    boolean_array = np.isin(label_array, labels_list)
    
    return boolean_array

def read_excel_to_lists(file_path, sheet_name=0):
    """Convert a pd dataframe to lists"""
    # Read the Excel file into a DataFrame without headers
    df = pd.read_excel(file_path, header=None, sheet_name=sheet_name)

    df = df.drop(0)

    # Initialize an empty list to store the lists of values
    data_lists = []

    # Iterate over each column in the DataFrame
    for column_name, column_data in df.items():
        # Convert the column values to a list and append to the data_lists
        data_lists.append(column_data.tolist())

    master_list = [[], [], []]


    for i in range(0, len(data_lists), 3):

        master_list[0].extend(data_lists[i])
        master_list[1].extend(data_lists[i+1])

        try:
            master_list[2].extend(data_lists[i+2])
        except IndexError:
            pass

    return master_list


def open_network(excel_file_path):
    """opens an unweighted network from the network excel file"""

    # Read the Excel file into a pandas DataFrame
    master_list = read_excel_to_lists(excel_file_path)

    # Create a graph
    G = nx.Graph()

    nodes_a = master_list[0]
    nodes_b = master_list[1]

    # Add edges to the graph
    for i in range(len(nodes_a)):
        G.add_edge(nodes_a[i], nodes_b[i])

    return G



def isolate_full_edges(nodes, edges, directory = None):
    """requires smart_dilate output to function properly"""

    print("Isolating full edges")

    if type(nodes) == str:
        nodes = tifffile.imread(nodes)

    if type(edges) == str:
        edges = tifffile.imread(edges)

    node_bools = binarize(nodes)

    del nodes

    real_edges = node_bools * edges

    del node_bools

    # Flatten the 3D array to a 1D array
    real_edges = real_edges.flatten()

    # Find unique values
    real_edges = np.unique(real_edges)

    edge_masks = labels_to_boolean(edges, real_edges)

    del real_edges

    edge_labels_1 = edge_masks * edges
     
    del edge_masks

    edge_labels_1 = binarize(edge_labels_1)

    edge_labels_1 = edge_labels_1 * 255

    if directory is None:

        tifffile.imsave(f"full_edges_for_component.tif", edge_labels_1)
        print("Full edge labels saved to full_edges_for_component.tif")
    else:
        tifffile.imsave(f"{directory}/full_edges_for_component.tif", edge_labels_1)
        print(f"Full edge labels saved to {directory}/full_edges_for_component.tif")


def isolate_edges(edges, network, iso_network, netlists = None, directory = None):

    print("Isolating edges")

    if netlists is None:

        master_list = read_excel_to_lists(network)
    else:
        master_list = netlists

    if directory is None:
        comp_list = read_excel_to_lists(iso_network)
    else:
        comp_list = read_excel_to_lists(f"{directory}/{iso_network}")

    edge_list = []

    node_1 = master_list[0]

    node_2 = master_list[1]

    edges_list = master_list[2]

    nodes_1 = comp_list[0]

    nodes_2 = comp_list[1]

    compare_list = []

    iso_list = []

    output_list = []

    for i in range(len(nodes_1)):
        item = [nodes_1[i], nodes_2[i]]
        iso_list.append(item)

    for i in range(len(node_1)):

        item = [node_1[i], node_2[i]]

        compare_list.append(item)

    for i in range(len(iso_list)):

        for k, item in enumerate(compare_list):
            if item == iso_list[i] or [item[1], item[0]] == iso_list[i]:
                add_item = [nodes_1[i], nodes_2[i], edges_list[k]]
                if add_item in output_list:
                    break
                else:
                    output_list.append(add_item)

                    edge_list.append(edges_list[k])

    # Convert to a DataFrame
    edges_df = pd.DataFrame(output_list, columns=["Node A", "Node B", "Edge C"])

    if directory is None:
        # Save to an Excel file
        edges_df.to_excel(f"{iso_network}", index=False)
        print(f"Isolated network file saved to {iso_network}")
    else:
        edges_df.to_excel(f"{directory}/{iso_network}", index=False)
        print(f"Isolated network file saved to {directory}/{iso_network}")


    mask2 = labels_to_boolean(edges, edge_list)

    mask2 = mask2 * edges

    # Convert boolean values to 0 and 255
    #mask2 = mask2.astype(np.uint8) * 255

    if directory is None:

        tifffile.imwrite(f"edges_for_{iso_network}.tif", mask2)
        print(f"Computational edge mask saved to edges_for_{iso_network}.tif")
    else:
        tifffile.imwrite(f"{directory}/edges_for_{iso_network}.tif", mask2)
        print(f"Computational edge mask saved to {directory}/edges_for_{iso_network}.tif")


    return output_list, mask2


def isolate_connected_component(nodes, excel, key=None, edge_file = None, netlists = None, search_region = None, directory = None):

    structure_3d = np.ones((3, 3, 3), dtype=int)

    if edge_file is not None and type(edge_file) == str:
        edge_file = tifffile.imread(edge_file)

    if type(nodes) == str:
        nodes = tifffile.imread(nodes)
        if len(np.unique(nodes)) == 2:
            nodes, num_nodes = ndimage.label(nodes, structure=structure_3d)

    if type(excel) == str:
        G = open_network(excel)
    else:
        G = excel

    if key is None:
        print("Isolating nodes of largest connected component")
        edges_df, mask, searchers = isolate_largest_connected(nodes, G, search_region = search_region, directory = directory)
        if edge_file is not None:
            output_list, mask2 = isolate_edges(edge_file, excel, 'largest_connected_component.xlsx', netlists = netlists, directory = directory)
            #isolate_full_edges(nodes, edge_file, 'largest_connected_component')
            return output_list, mask, mask2, searchers
        else:

            return edges_df, mask, searchers

    else:
        print("Isolating nodes of connected component containing specified key")
        edges_df, mask, searchers = isolate_key_connected(nodes, G, key, search_region = search_region, directory = directory)
        if edge_file is not None:
            output_list, mask2 = isolate_edges(edge_file, excel, f'connected_component_containing_{key}_node.xlsx', netlists = netlists, directory = directory)
            #isolate_full_edges(nodes, edge_file, 'connected_component_containing_specific_node')
            return output_list, mask, mask2, searchers
        else:
            return edges_df, mask, searchers


def _isolate_connected(G, key = None):

    if key is None:
        connected_components = list(nx.connected_components(G))
        Gcc = sorted(nx.connected_components(G), key=len, reverse=True)
        G0 = G.subgraph(Gcc[0])
        return G0

    else:
        # Get the connected component containing the specific node label
        connected_component = nx.node_connected_component(G, key)

        G0 = G.subgraph(connected_component)
        return G0



def isolate_largest_connected(masks, G, directory = None, search_region = None):
    # Read the Excel file into a pandas DataFrame

    # Get a list of connected components
    connected_components = list(nx.connected_components(G))
    Gcc = sorted(nx.connected_components(G), key=len, reverse=True)
    G0 = G.subgraph(Gcc[0])

    # Extract the edges of the largest connected component
    edges_largest_component = list(G0.edges)

    # Convert to a DataFrame
    edges_df = pd.DataFrame(edges_largest_component, columns=["Node A", "Node B"])

    if directory is None:
        # Save to an Excel file
        edges_df.to_excel("largest_connected_component.xlsx", index=False)
        print("Largest component nodes saved to largest_connected_component.xlsx")
    else:
        edges_df.to_excel(f"{directory}/largest_connected_component.xlsx", index=False)
        print(f"Largest component nodes saved to {directory}/largest_connected_component.xlsx")

    nodes_in_largest_component = list(G0)

    mask2 = labels_to_boolean(masks, nodes_in_largest_component)

    mask2 = mask2 * masks

    if search_region is not None:
        searchers = labels_to_boolean(search_region, nodes_in_largest_component)
        searchers = searchers * search_region
    else:
        searchers = None

    if directory is None:
        tifffile.imwrite("largest_connected_component.tif", mask2)
        print(f"Largest connected component image saved to largest_connected_component.tif")

    else:
        tifffile.imwrite(f"{directory}/largest_connected_component.tif", mask2)
        print(f"Largest connected component image saved to {directory}/largest_connected_component.tif")

    return edges_largest_component, mask2, searchers



def isolate_key_connected(masks, G, key, search_region = None, directory = None):


    # Get the connected component containing the specific node label
    connected_component = nx.node_connected_component(G, key)

    G0 = G.subgraph(connected_component)

    # Extract the edges of the largest connected component
    edges_component = list(G0.edges)

    # Convert to a DataFrame
    edges_df = pd.DataFrame(edges_component, columns=["Node A", "Node B"])


    if directory is None:
        # Save to an Excel file
        edges_df.to_excel(f"connected_component_containing_{key}_node.xlsx", index=False)
        print(f"Nodes containing {key} saved to connected_component_containing_{key}_node.xlsx")
    else:
        edges_df.to_excel(f"{directory}/connected_component_containing_{key}_node.xlsx", index=False)
        print(f"Nodes containing {key} saved to {directory}/connected_component_containing_{key}_node.xlsx")

    # Convert the set of nodes to a list
    nodes_in_component = list(connected_component)

    mask2 = labels_to_boolean(masks, nodes_in_component)

    mask2 = mask2 * masks

    # Convert boolean values to 0 and 255
    #mask2 = mask2.astype(np.uint8) * 255

    if search_region is not None:
        searchers = labels_to_boolean(search_region, nodes_in_component)
        searchers = searchers * search_region
    else:
        searchers = None

    if directory is None:
        tifffile.imwrite(f"connected_component_containing_{key}_node.tif", mask2)
        print(f"Connected component containing node {key} saved to connected_component_containing_{key}_node.tif")

    else:
        tifffile.imwrite(f"{directory}/connected_component_containing_{key}_node.tif", mask2)
        print(f"Connected component containing node {key} saved to {directory}/connected_component_containing_{key}_node.tif")

    return nodes_in_component, mask2, searchers


def extract_mothers(nodes, excel_file_path, centroid_dic = None, directory = None, louvain = True, ret_nodes = False, called = False):

    if type(nodes) == str:
        nodes = tifffile.imread(nodes)

        if np.unique(nodes) < 3:
            structure_3d = np.ones((3, 3, 3), dtype=int)
            nodes, num_nodes = ndimage.label(nodes, structure=structure_3d)

    if type(excel_file_path) == str:
        G, edge_weights = weighted_network(excel_file_path)
    else:
        G = excel_file_path

    if not called:

        if louvain:
            # Apply the Louvain algorithm for community detection
            partition = community_louvain.best_partition(G)
            some_communities = set(partition.values())
        else:
            some_communities = list(nx.community.label_propagation_communities(G))
            partition = {}
            for i, community in enumerate(some_communities):
                for node in community:
                    partition[node] = i + 1
    else:
        partition = louvain
        some_communities = (partition.keys())


    my_nodes, intercom_connections, connected_coms = get_border_nodes(partition, G)

    print(f"Number of intercommunity connections: {intercom_connections}")
    print(f"{len(connected_coms)} communities with any connectivity of {len(some_communities)} communities")

    mother_nodes = list(my_nodes)

    if ret_nodes or called:
        
        # Create a list to store nodes to be removed
        nodes_to_remove = []

        # Iterate through all nodes in the graph
        for node in G.nodes():
            # Check if the node's ID is not in the id_list
            if node not in mother_nodes:
                nodes_to_remove.append(node)

        # Remove the identified nodes from the graph
        G.remove_nodes_from(nodes_to_remove)

        if ret_nodes:

            return G


    if not ret_nodes:

        if centroid_dic is None:
            for item in nodes.shape:
                if item < 5:
                    down_factor = 1
                    break
                else:
                    down_factor = 5

            smalls2 = downsample(nodes, down_factor)

            centroid_dic = {}

            for item in mother_nodes:
                centroid = compute_centroid(smalls2, item)
                centroid_dic[item] = centroid

        mother_dict = {}


        for node in mother_nodes:
            mother_dict[node] = G.degree(node)

        #mask2 = labels_to_boolean(nodes, mother_nodes)

        smalls = labels_to_boolean(nodes, mother_nodes)

        if not called:

            # Convert boolean values to 0 and 255
            mask = smalls * nodes

            labels = node_draw.degree_draw(mother_dict, centroid_dic, smalls)

            # Convert dictionary to DataFrame with keys as index and values as a column
            df = pd.DataFrame.from_dict(mother_dict, orient='index', columns=['Degree'])

            # Rename the index to 'Node ID'
            df.index.name = 'Node ID'

            if directory is None:

                # Save DataFrame to Excel file
                df.to_excel('mothers.xlsx', engine='openpyxl')
                print("Mother list saved to mothers.xlsx")
            else:
                df.to_excel(f'{directory}/mothers.xlsx', engine='openpyxl')
                print(f"Mother list saved to {directory}/mothers.xlsx")

            if directory is None:

                tifffile.imwrite("mother_nodes.tif", mask)
                print("Mother nodes saved to mother_nodes.tif")
                tifffile.imwrite("mother_degree_labels.tif", labels)
                print(f"Mother degree labels saved to mother_degree_labels.tif")

            else:
                tifffile.imwrite(f"{directory}/mother_nodes.tif", mask)
                print(f"Mother nodes saved to {directory}/mother_nodes.tif")
                tifffile.imwrite(f"{directory}/mother_degree_labels.tif", labels)
                print(f"Mother degree labels saved to {directory}/mother_degree_labels.tif")


            smalls = node_draw.degree_infect(mother_dict, mask)

            if directory is None:

                tifffile.imwrite("mother_degree_labels_grayscale.tif", smalls)
                print("Mother graycale degree labels saved to mother_degree_labels_grayscale.tif")

            else:
                tifffile.imwrite(f"{directory}/mother_degree_labels_grayscale.tif", smalls)
                print(f"Mother graycale degree labels saved to {directory}/mother_degree_labels_grayscale.tif")


            return mother_nodes, smalls
        else:
            smalls = smalls * nodes
            return G, smalls







if __name__ == "__main__":

    # Read the Excel file into a pandas DataFrame
    excel_file_path = input("Excel file?: ")
    masks = input("watershedded, dilated glom mask?: ")
    masks = tifffile.imread(masks)
    masks = masks.astype(np.uint16)

    G = open_network(excel_file_path)

    # Get a list of connected components
    connected_components = list(nx.connected_components(G))

    largest_component = max(connected_components, key=len)


    # Choose a specific connected component (let's say, the first one)
    #selected_component = connected_components[0]

    # Convert the set of nodes to a list
    #nodes_in_component = list(selected_component)

    nodes_in_largest_component = list(largest_component)

    mask2 = labels_to_boolean(masks, nodes_in_largest_component)

    # Convert boolean values to 0 and 255
    mask2 = mask2.astype(np.uint8) * 255

    tifffile.imwrite("isolated_community.tif", mask2)